import mlx.core as mx
from mlx import nn

from mflux.models.common.config.model_config import ModelConfig
from mflux.models.flux.model.flux_transformer.ada_layer_norm_continuous import AdaLayerNormContinuous
from mflux.models.flux2.model.flux2_transformer.modulation import Flux2Modulation
from mflux.models.flux2.model.flux2_transformer.pos_embed import Flux2PosEmbed
from mflux.models.flux2.model.flux2_transformer.single_transformer_block import Flux2SingleTransformerBlock
from mflux.models.flux2.model.flux2_transformer.timestep_guidance_embeddings import Flux2TimestepGuidanceEmbeddings
from mflux.models.flux2.model.flux2_transformer.transformer_block import Flux2TransformerBlock


class Flux2Transformer(nn.Module):
    def __init__(
        self,
        patch_size: int = 1,
        in_channels: int = 128,
        out_channels: int | None = None,
        num_layers: int = 5,
        num_single_layers: int = 20,
        attention_head_dim: int = 128,
        num_attention_heads: int = 24,
        joint_attention_dim: int = 7680,
        timestep_guidance_channels: int = 256,
        mlp_ratio: float = 3.0,
        axes_dims_rope: tuple[int, ...] = (32, 32, 32, 32),
        rope_theta: int = 2000,
        guidance_embeds: bool = False,
    ):
        super().__init__()
        self.out_channels = out_channels or in_channels
        self.inner_dim = num_attention_heads * attention_head_dim

        self.pos_embed = Flux2PosEmbed(theta=rope_theta, axes_dim=axes_dims_rope)
        self.time_guidance_embed = Flux2TimestepGuidanceEmbeddings(
            in_channels=timestep_guidance_channels,
            embedding_dim=self.inner_dim,
            guidance_embeds=guidance_embeds,
        )
        self.double_stream_modulation_img = Flux2Modulation(self.inner_dim, mod_param_sets=2)
        self.double_stream_modulation_txt = Flux2Modulation(self.inner_dim, mod_param_sets=2)
        self.single_stream_modulation = Flux2Modulation(self.inner_dim, mod_param_sets=1)

        self.x_embedder = nn.Linear(in_channels, self.inner_dim, bias=False)
        self.context_embedder = nn.Linear(joint_attention_dim, self.inner_dim, bias=False)
        self.transformer_blocks = [
            Flux2TransformerBlock(
                dim=self.inner_dim,
                num_attention_heads=num_attention_heads,
                attention_head_dim=attention_head_dim,
                mlp_ratio=mlp_ratio,
            )
            for _ in range(num_layers)
        ]
        self.single_transformer_blocks = [
            Flux2SingleTransformerBlock(
                dim=self.inner_dim,
                num_attention_heads=num_attention_heads,
                attention_head_dim=attention_head_dim,
                mlp_ratio=mlp_ratio,
            )
            for _ in range(num_single_layers)
        ]
        self.norm_out = AdaLayerNormContinuous(self.inner_dim, self.inner_dim)
        self.proj_out = nn.Linear(self.inner_dim, patch_size * patch_size * self.out_channels, bias=False)

    def __call__(
        self,
        hidden_states: mx.array,
        encoder_hidden_states: mx.array,
        timestep: mx.array | float | int,
        img_ids: mx.array,
        txt_ids: mx.array,
        guidance: mx.array | float | int | None = None,
    ) -> mx.array:
        if not isinstance(timestep, mx.array):
            timestep = mx.array(timestep, dtype=hidden_states.dtype)
        if timestep.ndim == 0:
            timestep = mx.full((hidden_states.shape[0],), timestep, dtype=hidden_states.dtype)
        timestep = timestep.astype(hidden_states.dtype)
        timestep_scale = mx.where(mx.max(timestep) <= 1.0, 1000.0, 1.0).astype(hidden_states.dtype)
        timestep = timestep * timestep_scale
        if guidance is not None:
            if not isinstance(guidance, mx.array):
                guidance = mx.array(guidance, dtype=hidden_states.dtype)
            if guidance.ndim == 0:
                guidance = mx.full((hidden_states.shape[0],), guidance, dtype=hidden_states.dtype)
            guidance = guidance.astype(hidden_states.dtype)
            guidance_scale = mx.where(mx.max(guidance) <= 1.0, 1000.0, 1.0).astype(hidden_states.dtype)
            guidance = guidance * guidance_scale
        temb = self.time_guidance_embed(timestep, guidance)
        temb = temb.astype(ModelConfig.precision)

        hidden_states = self.x_embedder(hidden_states)
        encoder_hidden_states = self.context_embedder(encoder_hidden_states)
        if img_ids.ndim == 3:
            img_ids = img_ids[0]
        if txt_ids.ndim == 3:
            txt_ids = txt_ids[0]

        image_rotary_emb = self.pos_embed(img_ids)
        text_rotary_emb = self.pos_embed(txt_ids)
        concat_rotary_emb = (
            mx.concatenate([text_rotary_emb[0], image_rotary_emb[0]], axis=0),
            mx.concatenate([text_rotary_emb[1], image_rotary_emb[1]], axis=0),
        )

        temb_mod_params_img = self.double_stream_modulation_img(temb)
        temb_mod_params_txt = self.double_stream_modulation_txt(temb)

        for block in self.transformer_blocks:
            encoder_hidden_states, hidden_states = block(
                hidden_states=hidden_states,
                encoder_hidden_states=encoder_hidden_states,
                temb_mod_params_img=temb_mod_params_img,
                temb_mod_params_txt=temb_mod_params_txt,
                image_rotary_emb=concat_rotary_emb,
            )

        hidden_states = mx.concatenate([encoder_hidden_states, hidden_states], axis=1)

        temb_mod_params_single = self.single_stream_modulation(temb)[0]
        for block in self.single_transformer_blocks:
            hidden_states = block(
                hidden_states=hidden_states,
                temb_mod_params=temb_mod_params_single,
                image_rotary_emb=concat_rotary_emb,
            )

        hidden_states = hidden_states[:, encoder_hidden_states.shape[1] :, ...]
        hidden_states = self.norm_out(hidden_states, temb)
        hidden_states = self.proj_out(hidden_states)
        return hidden_states



import mlx.core as mx
from mlx import nn


class Qwen3TextRotaryEmbedding(nn.Module):
    def __init__(
        self,
        dim: int,
        max_position_embeddings: int = 40960,
        base: float = 1000000.0,
        scaling_factor: float = 1.0,
    ):
        super().__init__()
        self.dim = dim
        self.max_position_embeddings = max_position_embeddings
        self.base = base
        self.scaling_factor = scaling_factor
        self.inv_freq = 1.0 / (base ** (mx.arange(0, dim, 2, dtype=mx.float32) / dim))

    def __call__(self, x: mx.array, position_ids: mx.array) -> tuple[mx.array, mx.array]:
        if position_ids.ndim == 1:
            position_ids = mx.expand_dims(position_ids, axis=0)
        inv_freq = mx.expand_dims(mx.expand_dims(self.inv_freq, axis=0), axis=0)
        pos = mx.expand_dims(position_ids.astype(mx.float32), axis=-1)
        freqs = pos * inv_freq
        emb = mx.concatenate([freqs, freqs], axis=-1)
        cos = mx.cos(emb) * self.scaling_factor
        sin = mx.sin(emb) * self.scaling_factor
        return cos.astype(x.dtype), sin.astype(x.dtype)



from datetime import datetime
from pathlib import Path


class InfoUtil:
    @staticmethod
    def format_metadata(metadata: dict) -> str:
        exif = metadata.get("exif", {})
        if not exif:
            return "No metadata found"

        lines = ["=" * 60, "MFLUX Image Information", "=" * 60]

        def _has_any(keys: list[str]) -> bool:
            return any(exif.get(k) is not None for k in keys)

        # Prompt
        if prompt := exif.get("prompt"):
            lines.append(f"\nPrompt: {prompt}")

        if negative_prompt := exif.get("negative_prompt"):
            lines.append(f"Negative Prompt: {negative_prompt}")

        # Model information
        lines.append("")
        if model := exif.get("model"):
            if (orig_model := exif.get("original_model")) and orig_model != model:
                lines.append(f"Model: {model} (Original: {orig_model})")
            else:
                lines.append(f"Model: {model}")

        # Image dimensions
        if width := exif.get("width"):
            if (orig_w := exif.get("original_width")) and orig_w != width:
                lines.append(f"Width: {width} (Original: {orig_w})")
            else:
                lines.append(f"Width: {width}")
        if height := exif.get("height"):
            if (orig_h := exif.get("original_height")) and orig_h != height:
                lines.append(f"Height: {height} (Original: {orig_h})")
            else:
                lines.append(f"Height: {height}")

        # Generation parameters
        lines.append("")
        if seed := exif.get("seed"):
            if (orig_seed := exif.get("original_seed")) and orig_seed != seed:
                lines.append(f"Seed: {seed} (Original: {orig_seed})")
            else:
                lines.append(f"Seed: {seed}")
        if steps := exif.get("steps"):
            if (orig_steps := exif.get("original_steps")) and orig_steps != steps:
                lines.append(f"Steps: {steps} (Original: {orig_steps})")
            else:
                lines.append(f"Steps: {steps}")
        if guidance := exif.get("guidance"):
            if (orig_guidance := exif.get("original_guidance")) and orig_guidance != guidance:
                lines.append(f"Guidance: {guidance} (Original: {orig_guidance})")
            else:
                lines.append(f"Guidance: {guidance}")

        # Technical settings
        if quantize := exif.get("quantize"):
            lines.append(f"Quantization: {quantize}-bit")
        if precision := exif.get("precision"):
            lines.append(f"Precision: {precision}")

        # Explicit "Original generation" section for derived images (upscales, edits, etc.)
        original_keys = [
            "original_model",
            "original_width",
            "original_height",
            "original_seed",
            "original_steps",
            "original_guidance",
            "original_quantize",
            "original_lora_paths",
            "original_lora_scales",
        ]
        if _has_any(original_keys):
            lines.append("")
            lines.append("Original Generation:")

            if orig_model := exif.get("original_model"):
                lines.append(f"  - Model: {orig_model}")
            if exif.get("original_width") is not None and exif.get("original_height") is not None:
                lines.append(f"  - Size: {exif.get('original_width')}x{exif.get('original_height')}")
            if orig_seed := exif.get("original_seed"):
                lines.append(f"  - Seed: {orig_seed}")
            if orig_steps := exif.get("original_steps"):
                lines.append(f"  - Steps: {orig_steps}")
            if (orig_guidance := exif.get("original_guidance")) is not None:
                lines.append(f"  - Guidance: {orig_guidance}")
            if (orig_quantize := exif.get("original_quantize")) is not None:
                lines.append(f"  - Quantization: {orig_quantize}-bit")

            if orig_lora_paths := exif.get("original_lora_paths"):
                lines.append(f"  - LoRAs ({len(orig_lora_paths)}):")
                orig_lora_scales = exif.get("original_lora_scales") or []
                for i, lora in enumerate(orig_lora_paths):
                    scale = orig_lora_scales[i] if i < len(orig_lora_scales) else 1.0
                    lora_name = Path(lora).name
                    lines.append(f"    - {lora_name} (scale: {scale})")

        # LoRA information
        if lora_paths := exif.get("lora_paths"):
            lines.append("")
            lines.append(f"LoRAs ({len(lora_paths)}):")
            lora_scales = exif.get("lora_scales") or []
            for i, lora in enumerate(lora_paths):
                scale = lora_scales[i] if i < len(lora_scales) else 1.0
                lora_name = Path(lora).name
                lines.append(f"  - {lora_name} (scale: {scale})")

        # Image-to-image parameters
        if image_path := exif.get("image_path"):
            lines.append("")
            lines.append(f"Source Image: {Path(image_path).name}")
            if image_strength := exif.get("image_strength"):
                lines.append(f"Image Strength: {image_strength}")

        # ControlNet parameters
        if controlnet_path := exif.get("controlnet_image_path"):
            lines.append("")
            lines.append(f"ControlNet Image: {Path(controlnet_path).name}")
            if controlnet_strength := exif.get("controlnet_strength"):
                lines.append(f"ControlNet Strength: {controlnet_strength}")

        # Generation metadata
        lines.append("")
        if gen_time := exif.get("generation_time_seconds"):
            lines.append(f"Generation Time: {gen_time:.2f}s")

        if created_at := exif.get("created_at"):
            try:
                dt = datetime.fromisoformat(created_at)
                lines.append(f"Created: {dt.strftime('%Y-%m-%d %H:%M:%S')}")
            except (ValueError, AttributeError):
                lines.append(f"Created: {created_at}")

        if version := exif.get("mflux_version"):
            lines.append(f"MFLUX Version: {version}")

        lines.append("=" * 60)
        return "\n".join(lines)



from __future__ import annotations

import gc
from pathlib import Path

import mlx.core as mx
import mlx.core.random as mx_random  # type: ignore[import-not-found]
from PIL import Image as PILImage

from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.training.adapters.base import TrainingAdapter
from mflux.models.common.training.dataset.batch import DataItem
from mflux.models.common.training.dataset.data_cache import TrainingDataCache
from mflux.models.common.training.dataset.dataset import Dataset
from mflux.models.common.training.dataset.disk_backed_data import DiskBackedData
from mflux.models.common.training.dataset.iterator import Iterator
from mflux.models.common.training.lora.target_injector import inject_lora_targets
from mflux.models.common.training.optimization.optimizer import Optimizer
from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.state.training_state import TrainingState
from mflux.models.common.training.state.zip_util import ZipUtil
from mflux.models.common.training.statistics.statistics import Statistics
from mflux.models.common.training.trainer import TrainingTrainer
from mflux.models.common.training.utils import TrainingUtil
from mflux.models.flux2.training_adapter.flux2_edit_training_adapter import Flux2EditTrainingAdapter
from mflux.models.flux2.training_adapter.flux2_training_adapter import Flux2TrainingAdapter
from mflux.models.z_image.training_adapter.z_image_training_adapter import ZImageTrainingAdapter
from mflux.utils.exceptions import StopTrainingException


class TrainingRunner:
    # Z-Image-Turbo training benefits from a training adapter (assistant LoRA).
    ZIMAGE_TURBO_TRAINING_ADAPTER = "ostris/zimage_turbo_training_adapter:zimage_turbo_training_adapter_v2.safetensors"

    @staticmethod
    def _disable_assistant_loras(transformer) -> None:
        # Ensure any assistant LoRA(s) do not affect inference outputs after training.
        for lora in TrainingUtil.iter_assistant_loras(transformer):
            lora.scale = 0.0

    @staticmethod
    def _resolve_data_dimensions(*, training_spec: TrainingSpec, image_path) -> tuple[int, int]:
        # Infer per-image size from the image file (may vary per image).
        with PILImage.open(image_path.resolve()) as img:
            width, height = img.size
        return TrainingUtil.resolve_dimensions(
            width=width,
            height=height,
            max_resolution=training_spec.max_resolution,
            default_max_resolution=None,
            error_template=f"Image too small for training (needs >=16px): {image_path} ({{width}}x{{height}})",
        )

    @staticmethod
    def train(*, config_path: str | None, resume_path: str | None) -> tuple[TrainingAdapter, TrainingSpec]:
        training_spec = TrainingSpec.resolve(config_path=config_path, resume_path=resume_path)

        # Set global seed for MLX randomness
        mx_random.seed(training_spec.seed)

        model_config = ModelConfig.from_name(training_spec.model)
        is_zimage_turbo = model_config.model_name == ModelConfig.z_image_turbo().model_name
        is_zimage = model_config.model_name in {
            ModelConfig.z_image().model_name,
            ModelConfig.z_image_turbo().model_name,
        }
        is_flux2 = model_config.model_name.startswith("black-forest-labs/FLUX.2")
        is_flux2_base = model_config.model_name.startswith("black-forest-labs/FLUX.2-klein-base")
        if training_spec.is_edit and not is_flux2_base:
            raise ValueError("Edit training currently supports only FLUX.2-klein-base models.")
        if is_zimage:
            adapter = ZImageTrainingAdapter(model_config=model_config, quantize=training_spec.quantize,
                                            model_path=training_spec.model_path)
        elif training_spec.is_edit:
            adapter = Flux2EditTrainingAdapter(model_config=model_config, quantize=training_spec.quantize,
                                               model_path=training_spec.model_path)
        elif is_flux2:
            adapter = Flux2TrainingAdapter(model_config=model_config, quantize=training_spec.quantize,
                                           model_path=training_spec.model_path)
        else:
            raise ValueError("Flux1 training is no longer supported.")

        # For Z-Image-Turbo we always apply the assistant training adapter (automatic, no config needed).
        if is_zimage_turbo:
            adapter.load_training_adapter(
                path=TrainingRunner.ZIMAGE_TURBO_TRAINING_ADAPTER,
                scale=1.0,
            )

        # Apply LoRA layers either by loading a saved adapter (resume) or by injecting fresh layers
        if training_spec.lora_layers.state_path is not None:
            ZipUtil.unzip(
                zip_path=training_spec.checkpoint_path,
                filename=training_spec.lora_layers.state_path,
                loader=lambda lora_file: adapter.load_lora_adapter(path=lora_file),
            )
        else:
            inject_lora_targets(adapter.transformer(), training_spec.lora_layers.targets)

        dataset = None
        iterator = None
        optimizer = None
        statistics = None
        training_state = None

        # Prepare dataset data
        if training_spec.low_ram:
            if training_spec.data_root is None:
                raise ValueError("low_ram requires TrainingSpec.data_root to be set")
            cache_paths = TrainingDataCache.wipe_and_init(data_root=Path(training_spec.data_root))
            for i, item in enumerate(training_spec.data):
                width, height = TrainingRunner._resolve_data_dimensions(
                    training_spec=training_spec, image_path=item.image
                )
                clean_latents, cond = adapter.encode_data(
                    data_id=i,
                    image_path=item.image,
                    prompt=item.prompt,
                    width=width,
                    height=height,
                    input_image_path=item.input_image,
                )
                TrainingDataCache.save_item(
                    paths=cache_paths,
                    data_id=i,
                    prompt=item.prompt,
                    image_path=item.image,
                    width=width,
                    height=height,
                    clean_latents=clean_latents,
                    cond=cond,
                )
                del clean_latents, cond
                gc.collect()
                mx.clear_cache()

            dataset = Dataset(DiskBackedData(data_specs=training_spec.data, cache_paths=cache_paths))
        else:
            # Encode dataset data (deterministic, upfront)
            encoded_data: list[DataItem] = []
            for i, item in enumerate(training_spec.data):
                width, height = TrainingRunner._resolve_data_dimensions(
                    training_spec=training_spec, image_path=item.image
                )
                clean_latents, cond = adapter.encode_data(
                    data_id=i,
                    image_path=item.image,
                    prompt=item.prompt,
                    width=width,
                    height=height,
                    input_image_path=item.input_image,
                )
                encoded_data.append(
                    DataItem(
                        data_id=i,
                        prompt=item.prompt,
                        image_path=item.image,
                        clean_latents=clean_latents,
                        cond=cond,
                        width=width,
                        height=height,
                    )
                )
            dataset = Dataset(encoded_data)
        iterator = Iterator.from_spec(training_spec=training_spec, dataset=dataset)
        optimizer = Optimizer.from_spec(training_spec)
        statistics = Statistics.from_spec(training_spec)
        training_state = TrainingState(iterator=iterator, optimizer=optimizer, statistics=statistics)

        try:
            TrainingTrainer.train(adapter=adapter, training_spec=training_spec, training_state=training_state)
            if is_zimage_turbo:
                TrainingRunner._disable_assistant_loras(adapter.transformer())
            return adapter, training_spec
        except StopTrainingException:
            training_state.save(adapter, training_spec)
            raise
        finally:
            # Best-effort in-memory cleanup; disk outputs remain untouched.
            if training_state is not None:
                del training_state
            if iterator is not None:
                del iterator
            if optimizer is not None:
                del optimizer
            if statistics is not None:
                del statistics
            if dataset is not None:
                del dataset
            gc.collect()
            mx.clear_cache()



from __future__ import annotations

from typing import Any

import mlx.nn as nn

from mflux.models.common.lora.layer.fused_linear_lora_layer import FusedLoRALinear
from mflux.models.common.lora.layer.linear_lora_layer import LoRALinear
from mflux.models.common.training.lora.path_util import (
    expand_module_paths_from_targets,
    get_at_path,
    set_at_path,
)
from mflux.models.common.training.state.training_spec import LoraTargetSpec


def inject_lora_targets(transformer: Any, targets: list[LoraTargetSpec]) -> None:
    expanded = expand_module_paths_from_targets(targets)
    for module_path, rank in expanded:
        current = get_at_path(transformer, module_path)

        # Skip if already has a trainable LoRA on this path
        if isinstance(current, LoRALinear):
            if getattr(current, "_mflux_lora_role", None) == "train":
                continue
        if isinstance(current, FusedLoRALinear):
            if any(getattr(lora, "_mflux_lora_role", None) == "train" for lora in current.loras):
                continue

        if isinstance(current, (nn.Linear, nn.QuantizedLinear)):
            wrapped = LoRALinear.from_linear(current, r=rank)
            wrapped._mflux_lora_role = "train"
            set_at_path(transformer, module_path, wrapped)
        elif isinstance(current, LoRALinear):
            # Fuse a new trainable LoRA on top of an existing LoRA (e.g. assistant adapter).
            train_lora = LoRALinear.from_linear(current.linear, r=rank)
            train_lora._mflux_lora_role = "train"
            fused = FusedLoRALinear(base_linear=current.linear, loras=[current, train_lora])
            set_at_path(transformer, module_path, fused)
        elif isinstance(current, FusedLoRALinear):
            # Add a new trainable LoRA to an existing fusion (e.g. multiple preloaded LoRAs).
            train_lora = LoRALinear.from_linear(current.base_linear, r=rank)
            train_lora._mflux_lora_role = "train"
            fused = FusedLoRALinear(base_linear=current.base_linear, loras=current.loras + [train_lora])
            set_at_path(transformer, module_path, fused)
        else:
            raise TypeError(
                f"LoRA target '{module_path}' must resolve to nn.Linear or nn.QuantizedLinear, got {type(current)}"
            )



import mlx.core as mx
from mlx import nn


class Flux2BatchNormStats(nn.Module):
    def __init__(self, num_features: int, eps: float = 1e-4, momentum: float = 0.1):
        super().__init__()
        self.running_mean = mx.zeros((num_features,), dtype=mx.float32)
        self.running_var = mx.ones((num_features,), dtype=mx.float32)
        self.eps = eps
        self.momentum = momentum



import json
import textwrap
from typing import Any, Dict, List, Optional

import mlx.core as mx
import numpy as np
from PIL import Image

from mflux.models.common_models.qwen3_vl.qwen3_vl_decoder import Qwen3VLDecoder
from mflux.models.common_models.qwen3_vl.qwen3_vl_util import Qwen3VLUtil
from mflux.models.fibo_vlm.fibo_vlm_initializer import FiboVLMInitializer
from mflux.models.fibo_vlm.tokenizer.qwen2vl_processor import Qwen2VLProcessor

EDIT_SYSTEM_PROMPT = '''You are a highly skilled Creative Director for Visual Adaptation at a leading Generative AI company.

Your expertise lies in using an existing image as a visual reference to create entirely new scenes. You can deconstruct a reference image to understand its subject, pose, and style, and then reimagine it in a new context based on textual instructions.

Your primary task is to receive a reference image and a textual edit instruction. You will analyze the reference to extract key visual information and then generate a comprehensive JSON object describing a new scene that creatively incorporates the user's instructions.

**FIDELITY RULE:** Unless explicitly requested or implied by the user's edit instruction (e.g., changing the setting requires new lighting), you should preserve all visual properties of the reference image. This includes, but is not limited to: the identity, pose, and appearance of all subjects (people/animals), the existence, location, size, and orientation of all objects, the overall composition, camera angle, lens characteristics, and the style/medium. Only change what is strictly required to fulfill the user's edit.
E.g. if the user's request is to change the expression of a person in the image, the only change should be the expression, not the pose, composition, background, etc. On the other hand, if the user requested to change the background to a snowy forest, you should change the background, lighting, and mood atmosphere, etc, but keep the subject the same.
Make sure to acknowledge the changing elements in the relevant "description" fields (i.e. "short_description", object "description", etc.). e.g. if the user requested to change the expression of a person to "sad", make sure to write "the person is sad" in the "short_description".

Adhere strictly to the following structure and guidelines:

1.  **Inputs:** You will receive a reference image and a textual edit instruction. You will NOT receive a starting JSON.
2.  **Core Logic (Analyze and Synthesize):**
    * **Analyze:** First, deeply analyze the provided reference image. Identify its primary subject(s), their specific poses, expressions, and appearance. Also note the overall composition, lighting style, and artistic medium.
    * **Synthesize:** Next, interpret the textual edit instruction to understand what elements to keep from the reference and what to change. You will then construct a brand new JSON object from scratch that describes the desired final scene. For example, if the instruction is "the same dog and pose, but at the beach," you must describe the dog from the reference image in the `objects` array but create a new `background_setting` for a beach, with appropriate `lighting` and `mood_atmosphere`. **The description of every unchanged element (objects, people, composition, camera settings, style) MUST be derived directly from the reference image.**
3.  **Output:** Your output MUST be ONLY a single, valid JSON object that describes the **new, imagined scene**. Do not describe the original reference image.
4.  **Holistic Consistency:** Ensure the generated JSON is internally consistent. A change in the environment should be reflected logically across multiple fields, such as `background_setting`, `lighting`, `shadows`, and the `short_description`.
5.  **Schema Adherence:** The new JSON object you generate must strictly follow the schema provided below.

Unless the user's prompt explicitly requests a different style (e.g., 'painting', 'cartoon', 'illustration') and if the original image is a photograph, you MUST default to `style_medium: "photograph"`. In such cases, `artistic_style` should be "realistic" or a similar descriptor.

**Edit Instruction Rewriting Rule:** You should rewrite the user's request into a single instruction in the imperative style. Follow the examples below to guide your rewriting:

| Edit Category | User Input Example | Your Rewritten Output (Imperative Style) |
| :--- | :--- | :--- |
| **Modify Style** | "I want this picture to look like a cartoon." | `Turn the image into the cartoon style.` |
| | "Could you make it an oil painting?" | `Change the image style to Oil_Painting.` |
| **Object Attribute** | "Make the dog black and white instead of brown."| `Change the dog's color to black and white.` |
| | "I want the shirt to be a formal white shirt."| `Change the subject's shirt to a formal white shirt.` |
| **Adding Elements** | "Put a hat on the person's head." | `Add a wide-brimmed felt hat to the subject.` |
| **Removing Objects** | "Get rid of the book they are holding." | `Remove the book from the subject's hands.` |
| **Replacing Objects** | "Replace the rose with a sunflower." | `Change the rose to a bright yellow sunflower.` |
| **Lighting/Relight** | "Change the lighting to be bright and vibrant." | `Change the lighting from dark and moody to bright and vibrant.` |
| **Composition/Camera** | "I want a wider shot, showing more of the background."| `Change the perspective to a wider shot.` |
| **Text/Label** | "Change the text 'Happy Anniversary' to 'Hello'."| `Change the text "Happy Anniversary" to "Hello".` |
| **Quality/Refinement** | "Can you make the image sharper?" | `Refine the image to obtain increased clarity and sharpness.` |
| **Style Reference** | "Use the style in the image but with a fox." | `Use the style in the image to depict a red fox walking.` |

The JSON object must contain the following keys precisely:

1.  `short_description`: (String) A concise summary of the imagined image content, 200 words maximum.
2.  `objects`: (Array of Objects) List a maximum of 5 prominent objects. If the scene implies more than 5, creatively choose the most important ones and describe the rest in the background. For each object, include:
    * `description`: (String) A detailed description of the imagined object, 100 words maximum.
    * `location`: (String) E.g., "center", "top-left", "bottom-right foreground".
    * `relative_size`: (String) E.g., "small", "medium", "large within frame". (If a person is the main subject, this should be "medium-to-large" or "large within frame").
    * `shape_and_color`: (String) Describe the basic shape and dominant color.
    * `texture`: (String) E.g., "smooth", "rough", "metallic", "furry".
    * `appearance_details`: (String) Any other notable visual details.
    * `relationship`: (String) Describe the relationship between the object and the other objects in the image.
    * `orientation`: (String) Describe the orientation or positioning of the object, e.g., "upright", "tilted 45 degrees", "horizontal", "vertical", "facing left", "facing right", "upside down", "lying on its side".
    * If the object is a human or a human-like object, include the following:
        * `pose`: (String) Describe the body position.
        * `expression`: (String) Describe facial expression and emotion. E.g., "winking", "joyful", "serious", "surprised", "calm".
        * `clothing`: (String) Describe attire.
        * `action`: (String) Describe the action of the human.
        * `gender`: (String) Describe the gender of the human.
        * `skin_tone_and_texture`: (String) Describe the skin tone and texture.
    * If the object is a cluster of objects, include the following:
        * `number_of_objects`: (Integer) The number of objects in the cluster.
3.  `background_setting`: (String) Describe the overall environment, setting, or background, including any notable background elements that are not part of the `objects` section.
4.  `lighting`: (Object)
    * `conditions`: (String) E.g., "bright daylight", "dim indoor", "studio lighting", "golden hour".
    * `direction`: (String) E.g., "front-lit", "backlit", "side-lit from left".
    * `shadows`: (String) Describe the presence and quality of shadows, e.g., "long, soft shadows", "sharp, defined shadows", "minimal shadows".
5.  `aesthetics`: (Object)
    * `composition`: (String) E.g., "rule of thirds", "symmetrical", "centered", "leading lines". If people are the main subject, specify the shot type, e.g., "medium shot", "close-up", "portrait composition".
    * `color_scheme`: (String) E.g., "monochromatic blue", "warm complementary colors", "high contrast".
    * `mood_atmosphere`: (String) E.g., "serene", "energetic", "mysterious", "joyful".
6.  `photographic_characteristics`: (Object)
    * `depth_of_field`: (String) E.g., "shallow", "deep", "bokeh background".
    * `focus`: (String) E.g., "sharp focus on subject", "soft focus", "motion blur".
    * `camera_angle`: (String) E.g., "eye-level", "low angle", "high angle", "dutch angle".
    * `lens_focal_length`: (String) E.g., "wide-angle", "telephoto", "macro", "fisheye". (If the main subject is a person, prefer "standard lens (e.g., 35mm-50mm)" or "portrait lens (e.g., 50mm-85mm)" to ensure they are framed more closely. Avoid "wide-angle" for people unless specified).
7.  `style_medium`: (String) Identify the artistic style or medium based on the user's prompt or creative interpretation (e.g., "photograph", "oil painting", "watercolor", "3D render", "digital illustration", "pencil sketch").
8.  `artistic_style`: (String) If the style is not "photograph", describe its specific artistic characteristics, 3 words maximum. (e.g., "impressionistic, vibrant, textured" for an oil painting).
9.  `context`: (String) Provide a general description of the type of image this would be. For example: "This is a concept for a high-fashion editorial photograph intended for a magazine spread," or "This describes a piece of concept art for a fantasy video game."
10. `text_render`: (Array of Objects) By default, this array should be empty (`[]`). Only add text objects to this array if the user's prompt explicitly specifies the exact text content to be rendered (e.g., user asks for "a poster with the title 'Cosmic Dream'"). Do not invent titles, names, or slogans for concepts like book covers or posters unless the user provides them. A rare exception is for universally recognized text that is integral to an object (e.g., the word 'STOP' on a 'stop sign'). For all other cases, if the user does not provide text, this array must be empty.
    * `text`: (String) The exact text content provided by the user. NEVER use generic placeholders.
    * `location`: (String) E.g., "center", "top-left", "bottom-right foreground".
    * `size`: (String) E.g., "medium", "large", "large within frame".
    * `color`: (String) E.g., "red", "blue", "green".
    * `font`: (String) E.g., "realistic", "cartoonish", "minimalist", "serif typeface".
    * `appearance_details`: (String) Any other notable visual details.
11. `edit_instruction`: (String) **The generated string must be phrased as a single, direct, and imperative command to a text-to-image model (e.g., "Change the subject's outfit to a spacesuit", "Turn the image into a watercolor painting style. Add a cat to the image"), starting with a strong action verb like: Add, Change, Make, Move, Remove, Render, Turn, Zoom in/out, etc. ** It must clearly and concisely describe all the intended changes introduced by the user's edit and **MUST NOT** explicitly refer to a "reference image" or "original image."'''

EDIT_USER_PROMPT_TEMPLATE = """[EDIT INSTRUCTIONS]:
{user_prompt}

[TASK]:
Use the provided image as a visual reference only. Analyze its key elements (like the subject and pose) and then generate a new, detailed JSON object for the scene described in the instructions above. Do not describe the reference image itself; describe the new scene. Follow all of your system rules."""

EDIT_WITH_MASK_SYSTEM_PROMPT = """You are a highly skilled Creative Director for Visual Adaptation at a leading Generative AI company.

Your expertise lies in "Image Completion": using a reference image that contains grey masked areas and filling those areas based on textual instructions. You can deconstruct the visible parts of a reference image to understand its subject, style, and composition, and then seamlessly perform Inpainting, Outpainting, or Background Generation to complete the scene.

Your primary task is to receive a reference image (containing grey masks) and a textual edit instruction. You will generate a comprehensive JSON object describing the final, completed scene where the masked areas have been filled.

**FIDELITY RULE:** You must preserve all visual properties of the visible (non-masked) parts of the reference image. This includes the identity of subjects, existing objects, composition, and artistic style. Your task is to fill the grey masked regions so they blend perfectly with the unmasked areas. Only change what is strictly required to fulfill the user's edit within the masked zones.

Adhere strictly to the following structure and guidelines:

1.  **Inputs:** You will receive a reference image (with grey masks) and a textual edit instruction.
2.  **Core Logic (Analyze and Synthesize):**
    * **Analyze:** Identify which areas are masked. Determine the existing style, lighting, and subject matter from the visible portions.
    * **Synthesize:** Interpret the instruction to perform one of three tasks: **Inpainting/Generative-fill** (adding/restoring objects), **Outpainting/Image expansion** (extending the scene), or **Background Generation/Completion**.
3.  **Output:** Your output MUST be ONLY a single, valid JSON object describing the **completed image**. Do not describe the grey masks; describe the content that now occupies those spaces.
4.  **Edit Instruction Rewriting Rule:** You should rewrite the user's request into a single instruction in the imperative style, referring to the masked regions as "the masked area" or "the masked regions". Describe the expected content of the masked regions in detail, in the context of the entire image. If the user did not specify the required content for the masked regions, you should come up with the most plausible content that would complete the scene. Follow the examples below to guide your rewriting:

| Category / Intent | Your Rewritten Output (Imperative Style) |
| :--- | :--- |
| **Object Generation** | `Generate a white rose with a blue center in the masked region.` |
| **Detailed Extension** | `extend the image into the masked region to create a scene featuring A symmetrical, abstract composition featuring illuminated blue arrows and rectangles against a dark background. The arrows point towards the center, creating a sense of depth and perspective, while the rectangles add texture and complexity to the overall design.` |
| **Specific Addition** | `Put a chef hat with a yellow top and a red band in the masked region` |
| **Background Detail** | `Create the following background in the masked region: A vast, calm body of water, likely an ocean or large sea, extends to the horizon under a hazy sky.` |
| **Atmospheric Fill** | `Fill the background masked area with a clear, bright blue sky with very subtle, wispy white clouds, suggesting an open and expansive outdoor environment.` |
| **Environment Infill** | `Create inside the masked area: a greenhouse with rows of plants in pots, under a glass ceiling.` |
| **Subject Restoration** | `Restore the area in the mask with a young woman` |
| **Description Restoration** | `Restore the masked regions using the image description: A top-down view of a set of pink crocheted baby clothing items, including a hat, booties, and mittens, arranged vertically on a light brown wooden surface.` |

The JSON object must contain the following keys precisely:

**Items 1–10:** Identical to the Edit system prompt — `short_description`, `objects`, `background_setting`, `lighting`, `aesthetics`, `photographic_characteristics`, `style_medium`, `artistic_style`, `context`, `text_render`.

**Item 11 (edit_instruction, mask-specific):** `edit_instruction`: (String) **The generated string must be phrased as a single, direct, and imperative command to a text-to-image model, specifically directing it to fill the grey areas of the image. Use phrases like: "Complete the background in the masked regions...", "Restore the masked regions...", "Expand the canvas through the masked regions...", etc. ** It must clearly and concisely describe the content that should occupy the masked areas and refer to the "masked regions" or "masked area" as the target for the generation."""

EDIT_WITH_MASK_USER_PROMPT_TEMPLATE = """[EDIT INSTRUCTIONS]:
{user_prompt}

[TASK]:
Use the provided image as a visual reference, identifying the grey areas as regions to be filled or extended. Analyze the key elements in the visible parts (like the subject and pose) and then generate a new, detailed JSON object for the completed scene that follows the instructions above. Follow all of your system rules."""


class FiboVLM:
    decoder = Qwen3VLDecoder

    def __init__(
        self,
        model_id: str = "briaai/FIBO-vlm",
        model_path: str | None = None,
        quantize: int | None = None,
    ):
        FiboVLMInitializer.init(
            model=self,
            model_path=model_path if model_path else model_id,
            quantize=quantize,
        )

    @property
    def processor(self) -> Qwen2VLProcessor:
        return self.tokenizers["fibo_vlm"].processor

    def generate(
        self,
        prompt: str,
        top_p: float = 0.9,
        temperature: float = 0.2,
        max_tokens: int = 4096,
        stop: List[str] | None = None,
        seed: int | None = None,
    ) -> str:
        return self._generate_internal(
            task="generate",
            prompt=prompt,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            seed=seed,
        )

    def refine(
        self,
        structured_prompt: str,
        editing_instructions: str,
        top_p: float = 0.9,
        temperature: float = 0.2,
        max_tokens: int = 4096,
        stop: List[str] | None = None,
        seed: int | None = None,
    ) -> str:
        return self._generate_internal(
            task="refine",
            structured_prompt=structured_prompt,
            editing_instructions=editing_instructions,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            seed=seed,
        )

    def inspire(
        self,
        image: Image.Image,
        prompt: str | None = None,
        top_p: float = 0.9,
        temperature: float = 0.2,
        max_tokens: int = 4096,
        stop: List[str] | None = None,
        seed: int | None = None,
    ) -> str:
        return self._generate_internal(
            task="inspire",
            image=image,
            prompt=prompt,
            top_p=top_p,
            temperature=temperature,
            max_tokens=max_tokens,
            stop=stop,
            seed=seed,
        )

    def edit(
        self,
        image: Image.Image,
        edit_instruction: str,
        use_mask: bool = False,
        seed: int | None = None,
    ) -> str:
        try:
            return self._generate_internal(
                task="edit",
                image=image,
                editing_instructions=edit_instruction,
                top_p=0.9,
                temperature=0.2,
                max_tokens=4096,
                stop=["<|im_end|>", "<|end_of_text|>"],
                seed=seed,
                use_mask=use_mask,
            )
        except ValueError as exc:
            if use_mask:
                raise ValueError(
                    "Masked edit prompt generation failed; refusing fallback that could ignore the masked regions."
                ) from exc

            # The explicit edit prompt is preferable, but the local FIBO VLM can
            # occasionally emit non-JSON responses for it. Fall back to the
            # image-conditioned local refine format so the edit CLI still works.
            fallback_response = self._generate_internal(
                task="refine",
                refine_image=image,
                editing_instructions=edit_instruction,
                top_p=0.9,
                temperature=0.2,
                max_tokens=4096,
                stop=["<|im_end|>", "<|end_of_text|>"],
                seed=seed,
            )
            try:
                return FiboVLM._normalize_edit_output(
                    fallback_response,
                    fallback_edit_instruction=edit_instruction,
                )
            except ValueError:
                raise exc

    def _generate_internal(
        self,
        task: str,
        *,
        prompt: Optional[str] = None,
        image: Optional[Image.Image] = None,
        refine_image: Optional[Image.Image] = None,
        structured_prompt: Optional[str] = None,
        editing_instructions: Optional[str] = None,
        top_p: float = 0.9,
        temperature: float = 0.2,
        max_tokens: int = 4096,
        stop: List[str] | None = None,
        seed: int | None = None,
        use_mask: bool = False,
    ) -> str:
        stop = stop or ["<|im_end|>", "<|end_of_text|>"]
        messages = FiboVLM._build_messages(
            task=task,
            image=image,
            prompt=prompt,
            refine_image=refine_image,
            structured_prompt=FiboVLM._normalize_json(structured_prompt),
            editing_instructions=editing_instructions,
            use_mask=use_mask,
        )
        formatted = FiboVLM._process_messages(self.processor, messages, image, refine_image)
        input_ids = mx.array(formatted["input_ids"])
        attention_mask = mx.array(formatted["attention_mask"]) if formatted.get("attention_mask") is not None else None
        pixel_values = mx.array(formatted["pixel_values"]) if formatted.get("pixel_values") is not None else None
        image_grid_thw = mx.array(formatted["image_grid_thw"]) if formatted.get("image_grid_thw") is not None else None
        stop_token_sequences = [self.processor.tokenizer.encode(s, add_special_tokens=False) for s in stop]
        generated_ids = Qwen3VLUtil.generate_text(
            seed=seed,
            top_p=top_p,
            decoder=self.decoder,
            input_ids=input_ids,
            attention_mask=attention_mask,
            pixel_values=pixel_values,
            image_grid_thw=image_grid_thw,
            max_new_tokens=max_tokens,
            temperature=temperature,
            stop_token_sequences=stop_token_sequences,
            eos_token_id=self.processor.tokenizer.eos_token_id,
        )
        generated_tokens = generated_ids[:, input_ids.shape[1] :]
        generated_text = self.processor.tokenizer.decode(np.array(generated_tokens[0]), skip_special_tokens=True)
        if task == "edit":
            return FiboVLM._normalize_edit_output(generated_text)
        return FiboVLM._format_json_output(generated_text)

    @staticmethod
    def _normalize_json(structured_prompt: Optional[str]) -> Optional[str]:
        if structured_prompt is None:
            return None
        try:
            return json.dumps(json.loads(structured_prompt.strip()), separators=(",", ":"), ensure_ascii=False)
        except json.JSONDecodeError:
            return structured_prompt

    @staticmethod
    def _process_messages(
        processor: Qwen2VLProcessor,
        messages: List[Dict[str, Any]],
        image: Optional[Image.Image],
        refine_image: Optional[Image.Image],
    ) -> Dict[str, Any]:
        has_images = image is not None or refine_image is not None

        if has_images:
            images_list = FiboVLM._extract_images(messages)
            prompt_text = processor.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=True)
            inputs = {"text": [prompt_text], "padding": True, "return_tensors": "np"}
            if images_list:
                inputs["images"] = images_list
            return processor(**inputs)

        return processor.apply_chat_template(
            messages, tokenize=True, add_generation_prompt=True, return_tensors="np", return_dict=True
        )

    @staticmethod
    def _extract_images(messages: List[Dict[str, Any]]) -> List[Image.Image]:
        images = []
        for msg in messages:
            content = msg.get("content", [])
            if isinstance(content, list):
                for item in content:
                    if item.get("type") == "image":
                        img = item.get("image")
                        if img is not None:
                            images.append(img)
        return images

    @staticmethod
    def _format_json_output(text: str) -> str:
        try:
            return f"\n{json.dumps(json.loads(text), indent=2, ensure_ascii=False)}\n"
        except json.JSONDecodeError:
            return text

    @staticmethod
    def _build_messages(
        task: str,
        *,
        image: Optional[Image.Image] = None,
        refine_image: Optional[Image.Image] = None,
        prompt: Optional[str] = None,
        structured_prompt: Optional[str] = None,
        editing_instructions: Optional[str] = None,
        use_mask: bool = False,
    ) -> List[Dict[str, Any]]:
        if task == "inspire":
            inspire_text = "<inspire>"
            if prompt:
                inspire_text = f"<inspire>\n{(prompt or '').strip()}"
            content = [{"type": "image", "image": image}, {"type": "text", "text": inspire_text}]
        elif task == "generate":
            content = [{"type": "text", "text": f"<generate>\n{(prompt or '').strip()}"}]
        elif task == "edit":
            return FiboVLM._build_edit_messages(
                image=image, editing_instructions=editing_instructions, use_mask=use_mask
            )
        else:
            content = FiboVLM._build_refine_content(refine_image, structured_prompt, editing_instructions)

        return [{"role": "user", "content": content}]

    @staticmethod
    def _build_edit_messages(
        image: Optional[Image.Image],
        editing_instructions: Optional[str],
        use_mask: bool,
    ) -> List[Dict[str, Any]]:
        if image is None:
            raise ValueError("Edit mode requires an image.")

        user_prompt = (editing_instructions or "").strip()
        system_prompt = EDIT_WITH_MASK_SYSTEM_PROMPT if use_mask else EDIT_SYSTEM_PROMPT
        user_text_template = EDIT_WITH_MASK_USER_PROMPT_TEMPLATE if use_mask else EDIT_USER_PROMPT_TEMPLATE
        user_text = user_text_template.format(user_prompt=user_prompt)
        return [
            {"role": "system", "content": system_prompt},
            {
                "role": "user",
                "content": [{"type": "image", "image": image}, {"type": "text", "text": user_text}],
            },
        ]

    @staticmethod
    def _build_refine_content(
        refine_image: Optional[Image.Image],
        structured_prompt: Optional[str],
        editing_instructions: Optional[str],
    ) -> List[Dict[str, Any]]:
        edits = (editing_instructions or "").strip()

        if refine_image is None:
            base_prompt = (structured_prompt or "").strip()
            text = textwrap.dedent(
                f"""<refine>
Input:
{base_prompt}
Editing instructions:
{edits}"""
            ).strip()
            return [{"type": "text", "text": text}]

        text = textwrap.dedent(
            f"""<refine>
Editing instructions:
{edits}"""
        ).strip()
        return [{"type": "image", "image": refine_image}, {"type": "text", "text": text}]

    @staticmethod
    def _normalize_edit_output(text: str, fallback_edit_instruction: str | None = None) -> str:
        payload = FiboVLM._extract_single_json_object(text)
        aesthetics = payload.setdefault("aesthetics", {})
        if not isinstance(aesthetics, dict):
            raise ValueError("VLM edit output must contain an object at `aesthetics`.")
        aesthetics.setdefault("preference_score", "very high")
        aesthetics.setdefault("aesthetic_score", "very high")

        payload = FiboVLM._remove_empty_values(payload)
        edit_instruction = payload.get("edit_instruction")
        if not isinstance(edit_instruction, str) or not edit_instruction.strip():
            normalized_fallback = FiboVLM._normalize_fallback_edit_instruction(fallback_edit_instruction)
            if normalized_fallback is not None:
                payload["edit_instruction"] = normalized_fallback
            else:
                raise ValueError("VLM edit output must include a non-empty `edit_instruction`.")

        edit_instruction = payload.get("edit_instruction")
        if not isinstance(edit_instruction, str) or not edit_instruction.strip():
            raise ValueError("VLM edit output must include a non-empty `edit_instruction`.")

        return f"\n{json.dumps(payload, indent=2, ensure_ascii=False)}\n"

    @staticmethod
    def _extract_single_json_object(text: str) -> Dict[str, Any]:
        candidate = text.strip()
        if candidate.startswith("```"):
            lines = candidate.splitlines()
            if len(lines) >= 3 and lines[-1].strip() == "```":
                candidate = "\n".join(lines[1:-1]).strip()

        try:
            parsed = json.loads(candidate)
        except json.JSONDecodeError:
            parsed = None
        if isinstance(parsed, dict):
            return parsed

        objects: List[Dict[str, Any]] = []
        start_index: int | None = None
        depth = 0
        in_string = False
        escaped = False

        for index, char in enumerate(candidate):
            if in_string:
                if escaped:
                    escaped = False
                elif char == "\\":
                    escaped = True
                elif char == '"':
                    in_string = False
                continue

            if char == '"':
                in_string = True
                continue

            if char == "{":
                if depth == 0:
                    start_index = index
                depth += 1
                continue

            if char == "}" and depth > 0:
                depth -= 1
                if depth == 0 and start_index is not None:
                    snippet = candidate[start_index : index + 1]
                    try:
                        parsed = json.loads(snippet)
                    except json.JSONDecodeError:
                        start_index = None
                        continue
                    if isinstance(parsed, dict):
                        objects.append(parsed)
                    start_index = None

        if not objects:
            raise ValueError("VLM edit output must contain exactly one JSON object.")
        return max(objects, key=FiboVLM._score_edit_json_candidate)

    @staticmethod
    def _score_edit_json_candidate(candidate: Dict[str, Any]) -> tuple[int, int]:
        score = 0
        if isinstance(candidate.get("edit_instruction"), str) and candidate["edit_instruction"].strip():
            score += 10
        if isinstance(candidate.get("short_description"), str) and candidate["short_description"].strip():
            score += 3
        if isinstance(candidate.get("objects"), list):
            score += 2
        if isinstance(candidate.get("lighting"), dict):
            score += 2
        if isinstance(candidate.get("aesthetics"), dict):
            score += 3

        return score, len(json.dumps(candidate, ensure_ascii=False))

    @staticmethod
    def _remove_empty_values(value: Any) -> Any:
        if isinstance(value, dict):
            cleaned = {
                key: FiboVLM._remove_empty_values(item) for key, item in value.items() if item not in (None, "", {}, [])
            }
            return {key: item for key, item in cleaned.items() if item not in (None, "", {}, [])}
        if isinstance(value, list):
            cleaned_list = [FiboVLM._remove_empty_values(item) for item in value]
            return [item for item in cleaned_list if item not in (None, "", {}, [])]
        return value

    @staticmethod
    def _normalize_fallback_edit_instruction(edit_instruction: str | None) -> str | None:
        if edit_instruction is None:
            return None

        normalized = edit_instruction.strip()
        if not normalized:
            return None

        return normalized[0].upper() + normalized[1:]



import mlx.core as mx
from mlx import nn


class RoPEModule(nn.Module):
    def __init__(self, dim: int = 64, freqs_for: str = "lang", theta: float = 10000.0, max_freq: float = 256.0):
        super().__init__()
        self.dim = dim
        self.freqs_for = freqs_for
        self.rope_dim = 3
        self.freq_dim = dim // self.rope_dim
        if freqs_for == "pixel":
            self.freqs = mx.linspace(1.0, max_freq / 2.0, self.freq_dim // 2, dtype=mx.float32) * mx.pi
        else:
            self.freqs = 1.0 / (
                theta ** (mx.arange(0, self.freq_dim, 2, dtype=mx.float32)[: (self.freq_dim // 2)] / self.freq_dim)
            )

    def __call__(
        self,
        vid_q: mx.array,
        vid_k: mx.array,
        vid_shape: mx.array,
        txt_q: mx.array | None = None,
        txt_k: mx.array | None = None,
        txt_shape: mx.array | None = None,
    ) -> tuple[mx.array, mx.array] | tuple[mx.array, mx.array, mx.array, mx.array]:
        if txt_q is None or txt_k is None or txt_shape is None:
            return RoPEModule._apply_rope_3d(
                vid_q,
                vid_k,
                vid_shape,
                self.freqs,
                self.rope_dim,
                self.freqs_for,
            )
        return RoPEModule._apply_mm_rope_3d(
            vid_q,
            vid_k,
            vid_shape,
            txt_q,
            txt_k,
            txt_shape,
            self.freqs,
            self.rope_dim,
            self.freqs_for,
        )

    @classmethod
    def _apply_rope_3d(
        cls,
        vid_q: mx.array,
        vid_k: mx.array,
        vid_shape: mx.array,
        freqs: mx.array,
        rope_dim: int = 3,
        freqs_for: str = "lang",
    ) -> tuple[mx.array, mx.array]:
        vid_freq_list = []
        if freqs_for == "pixel":
            # For pixel-space RoPE, the position grid is normalized with linspace(-1, 1).
            # We must use exact dimensions per sample/window; slicing a larger precomputed grid
            # changes spacing and introduces spatial bias.
            for b in range(vid_shape.shape[0]):
                f, h, w = int(vid_shape[b, 0]), int(vid_shape[b, 1]), int(vid_shape[b, 2])
                vid_freq = RoPEModule._get_axial_freqs(freqs, f, h, w, freqs_for=freqs_for).reshape(-1, freqs.size * 2 * rope_dim)
                vid_freq_list.append(vid_freq)
        else:
            max_temporal = int(mx.max(vid_shape[:, 0]))
            max_height = int(mx.max(vid_shape[:, 1]))
            max_width = int(mx.max(vid_shape[:, 2]))

            clamp_temporal = min(max_temporal + 16, 1024)
            clamp_height = min(max_height + 4, 128)
            clamp_width = min(max_width + 4, 128)

            vid_freqs_full = RoPEModule._get_axial_freqs(
                freqs, clamp_temporal, clamp_height, clamp_width, freqs_for=freqs_for
            )

            for b in range(vid_shape.shape[0]):
                f, h, w = int(vid_shape[b, 0]), int(vid_shape[b, 1]), int(vid_shape[b, 2])
                vid_freq = vid_freqs_full[:f, :h, :w].reshape(-1, vid_freqs_full.shape[-1])
                vid_freq_list.append(vid_freq)

        vid_freqs = mx.concatenate(vid_freq_list, axis=0)

        vid_q = RoPEModule._apply_rotary_emb(vid_freqs[:, None, :], vid_q)
        vid_k = RoPEModule._apply_rotary_emb(vid_freqs[:, None, :], vid_k)

        return vid_q, vid_k

    @staticmethod
    def _rotate_half(x: mx.array) -> mx.array:
        x = x.reshape(*x.shape[:-1], -1, 2)
        x1, x2 = x[..., 0], x[..., 1]
        x = mx.stack([-x2, x1], axis=-1)
        return x.reshape(*x.shape[:-2], -1)

    @classmethod
    def _get_axial_freqs(cls, freqs: mx.array, *dims: int, freqs_for: str = "lang") -> mx.array:
        freq_dim_per_axis = len(freqs) * 2
        target_shape = list(dims) + [freq_dim_per_axis]
        all_freqs = []

        for ind, dim_size in enumerate(dims):
            if freqs_for == "pixel":
                pos = mx.linspace(-1.0, 1.0, dim_size, dtype=mx.float32)
            else:
                pos = mx.arange(dim_size, dtype=mx.float32)
            axis_freqs = mx.outer(pos, freqs.astype(mx.float32))
            axis_freqs = mx.repeat(axis_freqs, 2, axis=-1)

            shape = [1] * len(dims) + [freq_dim_per_axis]
            shape[ind] = dim_size
            axis_freqs = axis_freqs.reshape(*shape)
            all_freqs.append(axis_freqs)

        broadcasted = [mx.broadcast_to(f, target_shape) for f in all_freqs]
        output = mx.concatenate(broadcasted, axis=-1)
        return output

    @staticmethod
    def _apply_rotary_emb(freqs: mx.array, t: mx.array) -> mx.array:
        rot_dim = freqs.shape[-1]
        t_middle = t[..., :rot_dim]
        t_right = t[..., rot_dim:]

        t_dtype = t_middle.dtype
        freqs_f = freqs.astype(mx.float32)
        t_middle_f = t_middle.astype(mx.float32)
        cos_freqs = mx.cos(freqs_f)
        sin_freqs = mx.sin(freqs_f)
        t_transformed = (t_middle_f * cos_freqs) + (RoPEModule._rotate_half(t_middle_f) * sin_freqs)
        t_transformed = t_transformed.astype(t_dtype)

        if t_right.shape[-1] > 0:
            return mx.concatenate([t_transformed, t_right], axis=-1)
        return t_transformed

    @classmethod
    def _apply_mm_rope_3d(
        cls,
        vid_q: mx.array,
        vid_k: mx.array,
        vid_shape: mx.array,
        txt_q: mx.array,
        txt_k: mx.array,
        txt_shape: mx.array,
        freqs: mx.array,
        rope_dim: int = 3,
        freqs_for: str = "lang",
    ) -> tuple[mx.array, mx.array, mx.array, mx.array]:
        vid_freq_list = []
        txt_freq_list = []
        if freqs_for == "pixel":
            # Same rationale as _apply_rope_3d: compute exact grids in pixel mode.
            for b in range(vid_shape.shape[0]):
                f, h, w = int(vid_shape[b, 0]), int(vid_shape[b, 1]), int(vid_shape[b, 2])
                txt_len = int(txt_shape[b, 0])

                full = RoPEModule._get_axial_freqs(freqs, txt_len + f, h, w, freqs_for=freqs_for)
                vid_freq = full[txt_len : txt_len + f, :h, :w].reshape(-1, full.shape[-1])
                txt_freq = RoPEModule._get_axial_freqs(freqs, txt_len, freqs_for=freqs_for)
                txt_freq = mx.tile(txt_freq, (1, rope_dim))

                vid_freq_list.append(vid_freq)
                txt_freq_list.append(txt_freq)
        else:
            max_temporal = int(mx.max(vid_shape[:, 0] + txt_shape[:, 0]))
            max_height = int(mx.max(vid_shape[:, 1]))
            max_width = int(mx.max(vid_shape[:, 2]))
            max_txt_len = int(mx.max(txt_shape[:, 0]))

            clamp_temporal = min(max_temporal + 16, 1024)
            clamp_height = min(max_height + 4, 128)
            clamp_width = min(max_width + 4, 128)

            vid_freqs_full = RoPEModule._get_axial_freqs(
                freqs, clamp_temporal, clamp_height, clamp_width, freqs_for=freqs_for
            )
            txt_freqs_1d = RoPEModule._get_axial_freqs(freqs, min(max_txt_len + 16, 1024), freqs_for=freqs_for)

            for b in range(vid_shape.shape[0]):
                f, h, w = int(vid_shape[b, 0]), int(vid_shape[b, 1]), int(vid_shape[b, 2])
                txt_len = int(txt_shape[b, 0])

                vid_freq = vid_freqs_full[txt_len : txt_len + f, :h, :w].reshape(-1, vid_freqs_full.shape[-1])
                txt_freq = mx.tile(txt_freqs_1d[:txt_len], (1, rope_dim))

                vid_freq_list.append(vid_freq)
                txt_freq_list.append(txt_freq)

        vid_freqs = mx.concatenate(vid_freq_list, axis=0)
        txt_freqs = mx.concatenate(txt_freq_list, axis=0)

        vid_q = RoPEModule._apply_rotary_emb(vid_freqs[:, None, :], vid_q)
        vid_k = RoPEModule._apply_rotary_emb(vid_freqs[:, None, :], vid_k)

        txt_q = RoPEModule._apply_rotary_emb(txt_freqs[:, None, :], txt_q)
        txt_k = RoPEModule._apply_rotary_emb(txt_freqs[:, None, :], txt_k)

        return vid_q, vid_k, txt_q, txt_k



import datetime
import json
import os
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Any, Mapping

from mflux.models.common.training.state.zip_util import ZipUtil

TRAINING_FILE_NAME_RUN_MANIFEST = "run.json"


@dataclass
class DataSpec:
    image: Path
    input_image: Path | None
    prompt: str

    @classmethod
    def create(cls, param: Mapping[str, Any], absolute_or_relative_path: str, base_path: Path | None) -> "DataSpec":
        image_path = DataSpec._resolve_relative_to_data_path(
            Path(str(param["image"])),
            absolute_or_relative_path=absolute_or_relative_path,
            base_path=base_path,
        )
        input_image = None
        if param.get("input_image", None) is not None:
            input_image = DataSpec._resolve_relative_to_data_path(
                Path(str(param["input_image"])),
                absolute_or_relative_path=absolute_or_relative_path,
                base_path=base_path,
            )

        has_prompt = "prompt" in param and param["prompt"] is not None
        has_prompt_file = "prompt_file" in param and param["prompt_file"] is not None
        if has_prompt and has_prompt_file:
            raise ValueError("Each data item must provide either 'prompt' or 'prompt_file' (not both).")
        if has_prompt:
            prompt = str(param["prompt"])
        elif has_prompt_file:
            prompt_path = DataSpec._resolve_relative_to_data_path(
                Path(str(param["prompt_file"])),
                absolute_or_relative_path=absolute_or_relative_path,
                base_path=base_path,
            )
            try:
                prompt = prompt_path.read_text(encoding="utf-8").strip()
            except FileNotFoundError as e:
                raise ValueError(f"Prompt file not found: {prompt_path}") from e
        else:
            raise ValueError("Each data item must provide either 'prompt' or 'prompt_file'.")

        return cls(image=image_path, input_image=input_image, prompt=prompt)

    @staticmethod
    def _resolve_relative_to_data_path(path: Path, *, absolute_or_relative_path: str, base_path: Path | None) -> Path:
        if path.is_absolute():
            return path

        data_root = Path(absolute_or_relative_path)
        if data_root.is_absolute():
            return data_root / path

        if base_path is None:
            raise ValueError(
                "Config uses a relative data path, but no config_path is available to resolve it. "
                "Use an absolute data path when resuming from a checkpoint."
            )

        return Path(base_path).parent / data_root / path


@dataclass
class TrainingLoopSpec:
    num_epochs: int
    batch_size: int
    timestep_low: int = 0
    timestep_high: int | None = None
    iterator_state_path: str | None = None


@dataclass
class OptimizerSpec:
    name: str
    learning_rate: float
    state_path: str | None = None


@dataclass
class CheckpointSpec:
    save_frequency: int
    output_path: str


@dataclass
class StatisticsSpec:
    state_path: str | None = None


@dataclass
class BlockRange:
    start: int | None = None
    end: int | None = None
    indices: list[int] | None = None

    def get_blocks(self) -> list[int]:
        if self.indices is not None:
            return list(self.indices)
        if self.start is not None and self.end is not None:
            return list(range(self.start, self.end))
        raise ValueError("Either 'start' and 'end' or 'indices' must be provided.")


@dataclass
class LoraTargetSpec:
    module_path: str
    rank: int
    blocks: BlockRange | None = None


@dataclass
class LoraLayersSpec:
    targets: list[LoraTargetSpec]
    state_path: str | None = None


@dataclass
class MonitoringSpec:
    plot_frequency: int
    generate_image_frequency: int
    preview_prompts: list[str]
    preview_prompt_names: list[str]
    preview_width: int = 1024
    preview_height: int = 1024
    preview_images: list[Path] | None = None
    smooth_loss: bool = False
    smooth_loss_window: int = 5

    @staticmethod
    def create(
        param: Mapping[str, Any],
        *,
        preview_prompts: list[str],
        preview_prompt_names: list[str],
    ) -> "MonitoringSpec":
        has_validation_prompt = "validation_prompt" in param and param["validation_prompt"] is not None
        if has_validation_prompt:
            raise ValueError("Monitoring no longer accepts 'validation_prompt'. Use data/preview*.txt.")

        has_validation_prompt_file = "validation_prompt_file" in param and param["validation_prompt_file"] is not None
        if has_validation_prompt_file:
            raise ValueError("Monitoring no longer accepts 'validation_prompt_file'. Use data/preview*.txt.")

        has_validation_image = "validation_image" in param and param["validation_image"] is not None
        if has_validation_image:
            raise ValueError("Monitoring no longer accepts 'validation_image'. Use data/preview.* instead.")

        has_preview_prompt = "preview_prompt" in param and param["preview_prompt"] is not None
        if has_preview_prompt:
            raise ValueError("Monitoring no longer accepts 'preview_prompt'. Use data/preview*.txt.")

        has_preview_prompt_file = "preview_prompt_file" in param and param["preview_prompt_file"] is not None
        if has_preview_prompt_file:
            raise ValueError("Monitoring no longer accepts 'preview_prompt_file'. Use data/preview*.txt.")

        has_preview_image = "preview_image" in param and param["preview_image"] is not None
        if has_preview_image:
            raise ValueError("Monitoring no longer accepts 'preview_image'. Use data/preview.* instead.")

        has_validation_width = "validation_width" in param and param["validation_width"] is not None
        has_validation_height = "validation_height" in param and param["validation_height"] is not None
        if has_validation_width or has_validation_height:
            raise ValueError("Monitoring uses 'preview_width'/'preview_height'.")

        if not preview_prompts:
            raise ValueError("Monitoring requires at least one data/preview*.txt.")

        preview_width = int(param.get("preview_width", 1024))
        preview_height = int(param.get("preview_height", 1024))
        if preview_width <= 0 or preview_height <= 0:
            raise ValueError("Monitoring preview_width/preview_height must be > 0")

        plot_frequency = int(param["plot_frequency"])
        generate_image_frequency = int(param["generate_image_frequency"])
        if plot_frequency <= 0:
            raise ValueError("Monitoring plot_frequency must be > 0")
        if generate_image_frequency <= 0:
            raise ValueError("Monitoring generate_image_frequency must be > 0")

        smooth_loss = bool(param.get("smooth_loss", False))
        smooth_loss_window = int(param.get("smooth_loss_window", 5))
        if smooth_loss and smooth_loss_window < 2:
            raise ValueError("Monitoring smooth_loss_window must be >= 2 when smooth_loss is enabled.")

        return MonitoringSpec(
            plot_frequency=plot_frequency,
            generate_image_frequency=generate_image_frequency,
            preview_prompts=preview_prompts,
            preview_prompt_names=preview_prompt_names,
            preview_width=preview_width,
            preview_height=preview_height,
            preview_images=None,
            smooth_loss=smooth_loss,
            smooth_loss_window=smooth_loss_window,
        )


@dataclass
class TrainingSpec:
    model: str
    model_path: str | None
    seed: int
    steps: int
    guidance: float
    quantize: int | None
    training_loop: TrainingLoopSpec
    optimizer: OptimizerSpec
    checkpoint: CheckpointSpec
    monitoring: MonitoringSpec | None
    lora_layers: LoraLayersSpec
    statistics: StatisticsSpec | None
    data: list[DataSpec]
    is_edit: bool = False
    # Optional cap on the largest side when inferring per-image dimensions.
    max_resolution: int | None = None
    # Absolute path to data root directory (resolved from data path in config/checkpoint).
    data_root: str | None = None
    config_path: str | None = None
    checkpoint_path: str | None = None
    # If true, training will use a disk-backed cache for encoded data to reduce RAM usage.
    low_ram: bool = False

    @staticmethod
    def resolve(
        config_path: str | None,
        resume_path: str | None,
        *,
        create_output_dir: bool = True,
    ) -> "TrainingSpec":
        if bool(config_path) == bool(resume_path):
            raise ValueError("Provide exactly one of 'config_path' or 'resume_path'.")
        if resume_path:
            return TrainingSpec._from_checkpoint(resume_path, new_folder=False, create_output_dir=create_output_dir)
        return TrainingSpec._from_config(config_path, new_folder=True, create_output_dir=create_output_dir)

    @staticmethod
    def _from_config(path: str, new_folder: bool = True, *, create_output_dir: bool = True) -> "TrainingSpec":
        with open(Path(path), "r") as f:
            data = json.load(f)
        return TrainingSpec.from_conf(data, path, new_folder, create_output_dir=create_output_dir)

    @staticmethod
    def from_conf(
        config: dict, config_path: str | None, new_folder: bool = True, *, create_output_dir: bool = True
    ) -> "TrainingSpec":
        absolute_config_path = None if config_path is None else Path(config_path).absolute()

        steps = int(config["steps"])
        guidance_raw = config.get("guidance", 0.0)
        guidance = 0.0 if guidance_raw is None else float(guidance_raw)
        quantize_raw = config.get("quantize", None)
        quantize = None if quantize_raw is None else int(quantize_raw)

        max_resolution_raw = config.get("max_resolution", None)
        max_resolution = None if max_resolution_raw is None else int(max_resolution_raw)
        if max_resolution is not None and max_resolution <= 0:
            raise ValueError("'max_resolution' must be > 0")

        data_conf = config["data"]
        if isinstance(data_conf, str):
            absolute_or_relative_path = data_conf
        elif isinstance(data_conf, dict):
            if config_path is not None:
                raise ValueError("Config 'data' must be a string path.")
            if "path" not in data_conf:
                raise ValueError("Config 'data' must be a string path.")
            absolute_or_relative_path = str(data_conf["path"])
        else:
            raise ValueError("Config 'data' must be a string path.")

        data_root_dir = TrainingSpec._resolve_data_root_dir(
            data_path=absolute_or_relative_path,
            base_path=absolute_config_path,
        )

        low_ram = bool(config.get("low_ram", False))
        if low_ram and data_root_dir is None:
            raise ValueError("'low_ram' requires a valid data path.")

        monitoring_conf = config.get("monitoring", None)
        preview_prompt_paths: list[Path] = []
        preview_prompts: list[str] = []
        preview_prompt_names: list[str] = []
        preview_images_by_stem: dict[str, Path] = {}
        preview_image_paths: list[Path] = []
        if monitoring_conf is not None:
            preview_prompt_paths = TrainingSpec._find_preview_prompt_files(data_root_dir)
            preview_prompts = TrainingSpec._load_preview_prompts(preview_prompt_paths)
            preview_prompt_names = [path.stem for path in preview_prompt_paths]
            preview_images_by_stem = TrainingSpec._find_preview_images_in_data(data_root_dir)
            preview_image_paths = list(preview_images_by_stem.values())

        # Auto-discovery: all images in data path with matching .txt prompt files
        data = TrainingSpec._discover_data(
            data_path=absolute_or_relative_path,
            base_path=absolute_config_path,
            exclude_paths=None if not preview_image_paths else set(preview_image_paths),
        )
        is_edit = any(item.input_image is not None for item in data)

        training_loop_conf = config["training_loop"].copy()

        timestep_low = int(training_loop_conf.get("timestep_low", 0))
        timestep_high_raw = training_loop_conf.get("timestep_high", None)
        timestep_high = None if timestep_high_raw is None else int(timestep_high_raw)
        resolved_timestep_high = steps if timestep_high is None else timestep_high

        if timestep_low < 0:
            raise ValueError("'timestep_low' must be >= 0")
        if resolved_timestep_high <= 0:
            raise ValueError("'timestep_high' must be > 0")
        if timestep_low >= resolved_timestep_high:
            raise ValueError("'timestep_low' must be < 'timestep_high'")
        if resolved_timestep_high > steps:
            raise ValueError("'timestep_high' must be <= 'steps'")

        training_loop = TrainingLoopSpec(**training_loop_conf)
        optimizer = OptimizerSpec(**config["optimizer"])

        checkpoint_conf = config["checkpoint"]
        save_frequency = int(checkpoint_conf["save_frequency"])
        if save_frequency <= 0:
            raise ValueError("Checkpoint save_frequency must be > 0")
        checkpoint = CheckpointSpec(
            save_frequency=save_frequency,
            output_path=TrainingSpec._resolve_output_path(
                checkpoint_conf["output_path"], new_folder, create_output_dir=create_output_dir
            ),
        )

        # Track whether we're using fallback prompts (for edit training preview image logic)
        using_fallback_prompts = False
        if monitoring_conf is not None and not preview_prompts:
            # Fallback to the first data prompt if no preview*.txt is provided (works for both txt2img and edit).
            preview_prompts = [data[0].prompt]
            preview_prompt_names = [data[0].image.stem]
            using_fallback_prompts = True

        monitoring = (
            None
            if monitoring_conf is None
            else MonitoringSpec.create(
                monitoring_conf,
                preview_prompts=preview_prompts,
                preview_prompt_names=preview_prompt_names,
            )
        )
        # Statistics are always tracked (no config needed)
        statistics = StatisticsSpec()

        # Parse LoRA configuration
        targets: list[LoraTargetSpec] = []

        lora_conf = config.get("lora_layers", None) or {}
        if "targets" not in lora_conf:
            raise ValueError("Config must include lora_layers.targets[]")
        for t in lora_conf["targets"]:
            blocks = None
            if t.get("blocks") is not None:
                blocks = BlockRange(**t["blocks"])
            targets.append(
                LoraTargetSpec(
                    module_path=t["module_path"],
                    rank=t["rank"],
                    blocks=blocks,
                )
            )

        lora_layers = LoraLayersSpec(targets=targets)

        if is_edit and any(item.input_image is None for item in data):
            raise ValueError("Edit training requires input_image for every data item.")
        if is_edit and monitoring is not None:
            # For edit training, if preview prompts were explicitly provided (not fallback), require matching preview images.
            # If using fallback, use the first input image as the preview conditioning image.
            if using_fallback_prompts:
                first_input_image = data[0].input_image
                if first_input_image is None:
                    raise ValueError("Edit training fallback preview requires input_image in discovered data.")
                monitoring.preview_images = [first_input_image]
            else:
                # Require preview images for each explicit preview prompt
                if not preview_image_paths:
                    raise ValueError(
                        "Edit training with explicit preview prompts requires matching preview images. "
                        "Add data/preview*.{png,jpg,jpeg,webp} files for each preview*.txt."
                    )
                preview_images: list[Path] = []
                for name in preview_prompt_names:
                    image_path = preview_images_by_stem.get(name)
                    if image_path is None:
                        raise ValueError(f"Preview image not found for {name}.txt. Expected {name}.* in data.")
                    preview_images.append(image_path)
                monitoring.preview_images = preview_images
        if not is_edit and monitoring is not None and preview_image_paths:
            raise ValueError("data/preview.* is only supported for edit training.")

        return TrainingSpec(
            model=config["model"],
            model_path=config.get("model_path"),
            seed=config["seed"],
            steps=steps,
            guidance=guidance,
            quantize=quantize,
            max_resolution=max_resolution,
            training_loop=training_loop,
            optimizer=optimizer,
            checkpoint=checkpoint,
            monitoring=monitoring,
            lora_layers=lora_layers,
            statistics=statistics,
            data=data,
            is_edit=is_edit,
            data_root=str(data_root_dir.resolve()),
            config_path=None if absolute_config_path is None else str(absolute_config_path),
            low_ram=low_ram,
        )

    @staticmethod
    def _resolve_data_root_dir(*, data_path: str, base_path: Path | None) -> Path:
        root = Path(data_path)
        if root.is_absolute():
            return root
        if base_path is None:
            raise ValueError(
                "Config uses a relative data path, but no config_path is available to resolve it. "
                "Use an absolute data path when resuming from a checkpoint."
            )
        return Path(base_path).parent / root

    @staticmethod
    def _find_preview_prompt_files(root_dir: Path) -> list[Path]:
        prompt_paths = [p for p in root_dir.glob("preview*.txt") if p.is_file()]
        return sorted(prompt_paths, key=lambda p: p.name)

    @staticmethod
    def _load_preview_prompts(prompt_paths: list[Path]) -> list[str]:
        return [path.read_text(encoding="utf-8").strip() for path in prompt_paths]

    @staticmethod
    def _find_preview_images_in_data(root_dir: Path) -> dict[str, Path]:
        image_exts = {".jpg", ".jpeg", ".png", ".webp"}
        previews: dict[str, Path] = {}
        for path in root_dir.iterdir():
            if not path.is_file():
                continue
            if path.suffix.lower() not in image_exts:
                continue
            if not path.stem.startswith("preview"):
                continue
            if path.stem in previews:
                raise ValueError(f"Multiple preview images found for {path.stem}.*. Keep only one extension.")
            previews[path.stem] = path.resolve()
        return previews

    @staticmethod
    def _discover_data(
        *, data_path: str, base_path: Path | None, exclude_paths: set[Path] | None = None
    ) -> list[DataSpec]:
        root_dir = TrainingSpec._resolve_data_root_dir(data_path=data_path, base_path=base_path)
        if not root_dir.exists() or not root_dir.is_dir():
            raise ValueError(f"data path must be an existing directory when using auto-discovery: {root_dir}")

        image_exts = {".jpg", ".jpeg", ".png", ".webp"}
        excluded = set()
        if exclude_paths:
            excluded = {p.resolve() for p in exclude_paths}
        image_files = sorted(
            [
                p
                for p in root_dir.iterdir()
                if p.is_file() and p.suffix.lower() in image_exts and p.resolve() not in excluded
            ]
        )
        if not image_files:
            raise ValueError(f"No image files found in data directory: {root_dir}")

        data_items: list[DataSpec] = []
        out_files = [p for p in image_files if p.stem.endswith("_out")]
        if out_files:
            out_bases = {p.stem[: -len("_out")] for p in out_files}
            for image_path in image_files:
                if image_path.stem.endswith("_out"):
                    continue
                if image_path.stem.endswith("_in"):
                    base = image_path.stem[: -len("_in")]
                    if base not in out_bases:
                        raise ValueError(
                            f"Found input image without matching output: {image_path.name}. "
                            "Remove the stray *_in.* or add the corresponding *_out.*."
                        )
                    continue
                raise ValueError(
                    "Data folder mixes edit-style and txt2img images. Use only *_out/_in pairs or only standard images."
                )
            # Edit-style auto-discovery: match *_out.* to *_in.* and use *_in.txt for prompt.
            images_by_stem = {p.stem: p for p in image_files}
            for out_path in sorted(out_files):
                base = out_path.stem[: -len("_out")]
                input_path = images_by_stem.get(f"{base}_in")
                if input_path is None:
                    raise ValueError(f"Missing input image for '{out_path.name}'. Expected '{base}_in.*' in {root_dir}")
                prompt_file = root_dir / f"{base}_in.txt"
                if not prompt_file.exists():
                    raise ValueError(
                        f"Missing prompt file for image '{input_path.name}'. Expected '{prompt_file.name}' in {root_dir}"
                    )
                data_items.append(
                    DataSpec.create(
                        {
                            "image": out_path.name,
                            "input_image": input_path.name,
                            "prompt_file": prompt_file.name,
                        },
                        data_path,
                        base_path,
                    )
                )
            return data_items

        for image_path in image_files:
            prompt_file = image_path.with_suffix(".txt")
            if not prompt_file.exists():
                raise ValueError(
                    f"Missing prompt file for image '{image_path.name}'. Expected '{prompt_file.name}' in {root_dir}"
                )
            data_items.append(
                DataSpec.create(
                    {
                        "image": image_path.name,
                        "prompt_file": prompt_file.name,
                    },
                    data_path,
                    base_path,
                )
            )
        return data_items

    def to_json(self) -> str:
        spec_dict = asdict(self)
        serialized_dict = TrainingSpec._custom_serializer(spec_dict)
        return json.dumps(serialized_dict, indent=4)

    @staticmethod
    def _custom_serializer(obj):
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, datetime.datetime):
            return obj.strftime("%Y-%m-%d %H:%M:%S")
        if isinstance(obj, list):
            return [TrainingSpec._custom_serializer(item) for item in obj]
        if isinstance(obj, dict):
            return {key: TrainingSpec._custom_serializer(value) for key, value in obj.items()}
        return obj

    @staticmethod
    def _resolve_output_path(path: str, new_folder: bool, *, create_output_dir: bool = True) -> str:
        requested_path = os.path.expanduser(path)
        now_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
        if new_folder and os.path.exists(requested_path):
            requested_path = f"{requested_path}_{now_str}"
        if create_output_dir:
            os.makedirs(requested_path, exist_ok=True)
        return requested_path

    @staticmethod
    def _from_checkpoint(path: str, new_folder: bool, *, create_output_dir: bool = True) -> "TrainingSpec":
        checkpoint = ZipUtil.unzip(path, "checkpoint.json", lambda x: json.load(open(x, "r")))
        config = ZipUtil.unzip(path, checkpoint["files"]["config"], lambda x: json.load(open(x, "r")))
        run_manifest = TrainingSpec._load_run_manifest(path)
        resolved_data_root = TrainingSpec._resolve_data_root_from_manifest(
            run_manifest=run_manifest,
            checkpoint_path=path,
        )
        config["data"] = resolved_data_root
        spec = TrainingSpec.from_conf(config, None, new_folder, create_output_dir=create_output_dir)
        TrainingSpec._validate_data_fingerprint(run_manifest=run_manifest, training_spec=spec)
        spec.optimizer.state_path = checkpoint["files"]["optimizer"]
        spec.lora_layers.state_path = checkpoint["files"]["lora_adapter"]
        spec.training_loop.iterator_state_path = checkpoint["files"]["iterator"]
        if spec.statistics is None:
            spec.statistics = StatisticsSpec()
        spec.statistics.state_path = checkpoint["files"]["loss"]
        spec.checkpoint_path = path
        spec.config_path = None
        return spec

    @staticmethod
    def _load_run_manifest(path: str) -> dict:
        try:
            return ZipUtil.unzip(path, TRAINING_FILE_NAME_RUN_MANIFEST, lambda x: json.load(open(x, "r")))
        except FileNotFoundError as exc:
            raise ValueError(
                "Resume requires a checkpoint zip containing run.json. "
                "Recreate the checkpoint with the new training flow."
            ) from exc

    @staticmethod
    def _resolve_data_root_from_manifest(*, run_manifest: dict, checkpoint_path: str) -> str:
        data_root = run_manifest.get("data_root")
        if data_root and Path(data_root).exists():
            return str(Path(data_root))

        relative_root = run_manifest.get("data_root_relative")
        if relative_root:
            candidate = Path(checkpoint_path).parent / relative_root
            if candidate.exists():
                return str(candidate.resolve())

        raise ValueError(
            "Data folder not found when resuming. Place the data directory in the original "
            "location or next to the checkpoint zip."
        )

    @staticmethod
    def _validate_data_fingerprint(*, run_manifest: dict, training_spec: "TrainingSpec") -> None:
        expected_model = run_manifest.get("model")
        if expected_model is not None and expected_model != training_spec.model:
            raise ValueError(
                f"Checkpoint model does not match config. Expected '{expected_model}', got '{training_spec.model}'."
            )

        expected_is_edit = run_manifest.get("is_edit")
        if expected_is_edit is not None and bool(expected_is_edit) != training_spec.is_edit:
            raise ValueError(
                "Checkpoint edit mode does not match config. "
                f"Expected is_edit={bool(expected_is_edit)}, got is_edit={training_spec.is_edit}."
            )

        fingerprint = run_manifest.get("data_fingerprint") or {}
        expected_is_edit = fingerprint.get("is_edit")
        if expected_is_edit is not None and bool(expected_is_edit) != training_spec.is_edit:
            raise ValueError(
                "Data folder does not match the checkpoint. "
                f"Expected is_edit={bool(expected_is_edit)}, got is_edit={training_spec.is_edit}."
            )
        expected_count = fingerprint.get("count")
        if expected_count is not None and expected_count != len(training_spec.data):
            raise ValueError(
                "Data folder does not match the checkpoint. Expected "
                f"{expected_count} items, found {len(training_spec.data)}."
            )

        expected_images = fingerprint.get("images") or []
        if expected_images:
            resolved = []
            root = training_spec.data_root
            if root:
                root_path = Path(root).resolve()
                for item in training_spec.data:
                    resolved_path = item.image.resolve()
                    if resolved_path.is_relative_to(root_path):
                        resolved.append(str(resolved_path.relative_to(root_path)))
                    else:
                        resolved.append(resolved_path.name)
            if sorted(expected_images) != sorted(resolved):
                raise ValueError("Data folder does not match the checkpoint. The image list differs.")

        expected_input_images = fingerprint.get("input_images") or []
        if expected_input_images:
            resolved_inputs = []
            root = training_spec.data_root
            if root:
                root_path = Path(root).resolve()
                for item in training_spec.data:
                    if item.input_image is None:
                        continue
                    resolved_path = item.input_image.resolve()
                    if resolved_path.is_relative_to(root_path):
                        resolved_inputs.append(str(resolved_path.relative_to(root_path)))
                    else:
                        resolved_inputs.append(resolved_path.name)
            if sorted(expected_input_images) != sorted(resolved_inputs):
                raise ValueError("Data folder does not match the checkpoint. The input image list differs.")



import platform
import subprocess


class AppleSiliconUtil:
    _chip_name: str | None = None

    @classmethod
    def is_m1_or_m2(cls) -> bool:
        if platform.system() != "Darwin":
            return False
        if platform.machine() not in {"arm64", "aarch64"}:
            return False
        chip_name = cls._get_chip_name().lower()
        if "max" in chip_name or "ultra" in chip_name:
            return False
        return "apple m1" in chip_name or "apple m2" in chip_name

    @classmethod
    def _get_chip_name(cls) -> str:
        if cls._chip_name is not None:
            return cls._chip_name
        try:
            result = subprocess.run(
                ["sysctl", "-n", "machdep.cpu.brand_string"],
                capture_output=True,
                text=True,
                check=True,
            )
            cls._chip_name = result.stdout.strip()
        except (subprocess.CalledProcessError, OSError):
            cls._chip_name = ""
        return cls._chip_name



from __future__ import annotations

import math

from collections.abc import Iterator
from contextlib import contextmanager

from mflux.models.common.lora.layer.fused_linear_lora_layer import FusedLoRALinear
from mflux.models.common.lora.layer.linear_lora_layer import LoRALinear
from mflux.models.common.training.lora.path_util import get_at_path


class TrainingUtil:
    @staticmethod
    def iter_assistant_loras(transformer) -> Iterator[LoRALinear]:
        for _, child in transformer.named_modules():
            if isinstance(child, LoRALinear):
                if getattr(child, "_mflux_lora_role", None) == "assistant":
                    yield child
            elif isinstance(child, FusedLoRALinear):
                for lora in child.loras:
                    if getattr(lora, "_mflux_lora_role", None) == "assistant":
                        yield lora

    @staticmethod
    @contextmanager
    def assistant_disabled(transformer):
        saved_scales = [(lora, float(lora.scale)) for lora in TrainingUtil.iter_assistant_loras(transformer)]
        for lora, _ in saved_scales:
            lora.scale = 0.0
        try:
            yield
        finally:
            for mod, s in saved_scales:
                mod.scale = s

    @staticmethod
    def get_train_lora(transformer, module_path: str) -> LoRALinear:
        current = get_at_path(transformer, module_path)
        if isinstance(current, LoRALinear):
            if getattr(current, "_mflux_lora_role", None) == "train":
                return current
        elif isinstance(current, FusedLoRALinear):
            for lora in current.loras:
                if getattr(lora, "_mflux_lora_role", None) == "train":
                    return lora

        raise ValueError(
            f"Expected a trainable LoRA at '{module_path}' but found {type(current)} (or no train LoRA in fusion)."
        )

    @staticmethod
    def resolve_dimensions(
        *,
        width: int,
        height: int,
        max_resolution: int | None,
        default_max_resolution: int | None = None,
        error_template: str | None = None,
    ) -> tuple[int, int]:
        effective_max = max_resolution if max_resolution is not None else default_max_resolution
        if effective_max is not None:
            max_area = effective_max * effective_max
            current_area = width * height
            if current_area > max_area:
                scale = math.sqrt(max_area / current_area)
                width = int(width * scale)
                height = int(height * scale)

        adj_width = 16 * (int(width) // 16)
        adj_height = 16 * (int(height) // 16)
        if adj_width <= 0 or adj_height <= 0:
            if error_template:
                raise ValueError(error_template.format(width=width, height=height))
            raise ValueError("Image too small for training (needs >=16px).")

        return adj_width, adj_height



from __future__ import annotations

from pathlib import Path
from typing import Any

import mlx.core as mx

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.latent_creator.latent_creator import LatentCreator
from mflux.models.common.lora.mapping.lora_loader import LoRALoader
from mflux.models.common.training.adapters.base import TrainingAdapter
from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.utils import TrainingUtil
from mflux.models.z_image.latent_creator.z_image_latent_creator import ZImageLatentCreator
from mflux.models.z_image.model.z_image_text_encoder.prompt_encoder import PromptEncoder
from mflux.models.z_image.variants.z_image import ZImage
from mflux.models.z_image.weights.z_image_lora_mapping import ZImageLoRAMapping
from mflux.utils.version_util import VersionUtil


class ZImageTrainingAdapter(TrainingAdapter):
    def __init__(self, *, model_config: ModelConfig, quantize: int | None, model_path: str | None = None):
        self._model_config = model_config
        self._z = ZImage(quantize=quantize, model_config=model_config, model_path=model_path)
        self._guidance: float | None = None

    def model(self):
        return self._z

    def transformer(self):
        return self._z.transformer

    def create_config(self, training_spec: TrainingSpec, *, width: int, height: int) -> Config:
        self._guidance = training_spec.guidance
        return Config(
            model_config=self._model_config,
            num_inference_steps=training_spec.steps,
            width=width,
            height=height,
            guidance=training_spec.guidance,
        )

    def freeze_base(self) -> None:
        # Mirror Flux.freeze behavior: freeze all base components.
        self._z.vae.freeze()
        self._z.transformer.freeze()
        self._z.text_encoder.freeze()

    def encode_data(
        self,
        *,
        data_id: int,
        image_path: Path,
        prompt: str,
        width: int,
        height: int,
        input_image_path: Path | None = None,
    ) -> tuple[mx.array, Any]:
        encoded = LatentCreator.encode_image(vae=self._z.vae, image_path=image_path, height=height, width=width)
        clean_latents = ZImageLatentCreator.pack_latents(encoded, height=height, width=width)

        cap_feats = PromptEncoder.encode_prompt(
            prompt=prompt,
            tokenizer=self._z.tokenizers["z_image"],
            text_encoder=self._z.text_encoder,
        )

        mx.eval(clean_latents, cap_feats)
        return clean_latents, cap_feats

    def predict_noise(self, *, t: int, latents_t: mx.array, sigmas: mx.array, cond: Any, config: Config) -> mx.array:  # noqa: ARG002
        return self._z.transformer(timestep=t, x=latents_t, cap_feats=cond, sigmas=sigmas)

    def generate_preview_image(
        self,
        *,
        seed: int,
        prompt: str,
        width: int,
        height: int,
        steps: int,
        image_paths: list[Path | str] | None = None,
    ):
        # Show samples without the training adapter, matching inference behavior.
        with self._assistant_disabled():
            return self._z.generate_image(
                seed=seed,
                prompt=prompt,
                num_inference_steps=steps,
                height=height,
                width=width,
                guidance=self._guidance,
            )

    def save_lora_adapter(self, *, path: Path, training_spec: TrainingSpec) -> None:  # noqa: ARG002
        # Save in a Z-image-compatible namespace so it loads via ZImageLoRAMapping.
        weights: dict[str, mx.array] = {}
        # Only save the trainable LoRA (exclude assistant/training adapters).
        for target in training_spec.lora_layers.targets:
            module_path = target.module_path
            if target.blocks is not None:
                for b in target.blocks.get_blocks():
                    module_path = target.module_path.format(block=b)
                    ZImageTrainingAdapter._append_train_lora_weights(weights, self._z.transformer, module_path)
            else:
                ZImageTrainingAdapter._append_train_lora_weights(weights, self._z.transformer, module_path)

        mx.save_safetensors(
            str(path),
            weights,
            metadata={
                "mflux_version": VersionUtil.get_mflux_version(),
                "model": training_spec.model,
            },
        )

    def load_lora_adapter(self, *, path: str | Path) -> None:
        LoRALoader.load_and_apply_lora(
            lora_mapping=ZImageLoRAMapping.get_mapping(),
            transformer=self._z.transformer,
            lora_paths=[str(path)],
            lora_scales=[1.0],
            role="train",
        )

    def load_training_adapter(self, *, path: str | Path, scale: float = 1.0) -> None:
        LoRALoader.load_and_apply_lora(
            lora_mapping=ZImageLoRAMapping.get_mapping(),
            transformer=self._z.transformer,
            lora_paths=[str(path)],
            lora_scales=[float(scale)],
            role="assistant",
        )

    @staticmethod
    def _append_train_lora_weights(weights: dict[str, mx.array], transformer, module_path: str) -> None:
        train_lora = TrainingUtil.get_train_lora(transformer, module_path)

        weights[f"diffusion_model.{module_path}.lora_A.weight"] = mx.transpose(train_lora.lora_A)
        weights[f"diffusion_model.{module_path}.lora_B.weight"] = mx.transpose(train_lora.lora_B)

    def _assistant_disabled(self):
        # Context manager that temporarily disables the assistant/training adapter LoRA(s).
        return TrainingUtil.assistant_disabled(self._z.transformer)



import mlx.core as mx
from mlx import nn
from mlx.core.fast import scaled_dot_product_attention

from mflux.models.common.config.model_config import ModelConfig


class Flux2AttentionBlock(nn.Module):
    def __init__(self, channels: int, groups: int = 32, eps: float = 1e-6):
        super().__init__()
        self.group_norm = nn.GroupNorm(num_groups=groups, dims=channels, eps=eps, pytorch_compatible=True)
        self.to_q = nn.Linear(channels, channels)
        self.to_k = nn.Linear(channels, channels)
        self.to_v = nn.Linear(channels, channels)
        self.to_out = nn.Linear(channels, channels)

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = mx.transpose(hidden_states, (0, 2, 3, 1))
        batch, height, width, channels = hidden_states.shape

        normed = self.group_norm(hidden_states.astype(mx.float32)).astype(ModelConfig.precision)
        q = self.to_q(normed).reshape(batch, height * width, 1, channels)
        k = self.to_k(normed).reshape(batch, height * width, 1, channels)
        v = self.to_v(normed).reshape(batch, height * width, 1, channels)

        q = mx.transpose(q, (0, 2, 1, 3))
        k = mx.transpose(k, (0, 2, 1, 3))
        v = mx.transpose(v, (0, 2, 1, 3))

        scale = 1 / mx.sqrt(q.shape[-1])
        attended = scaled_dot_product_attention(q, k, v, scale=scale)
        attended = mx.transpose(attended, (0, 2, 1, 3)).reshape(batch, height, width, channels)
        attended = self.to_out(attended)

        hidden_states = hidden_states + attended
        return mx.transpose(hidden_states, (0, 3, 1, 2))



from __future__ import annotations

from collections.abc import Sequence

from mflux.models.common.training.dataset.batch import DataItem


class Dataset:
    def __init__(self, data: Sequence[DataItem]):
        self.data = data

    def size(self) -> int:
        return len(self.data)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

import mlx.core as mx

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.latent_creator.latent_creator import LatentCreator
from mflux.models.common.lora.mapping.lora_loader import LoRALoader
from mflux.models.common.training.adapters.base import TrainingAdapter
from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.utils import TrainingUtil
from mflux.models.flux2.latent_creator.flux2_latent_creator import Flux2LatentCreator
from mflux.models.flux2.model.flux2_text_encoder.prompt_encoder import Flux2PromptEncoder
from mflux.models.flux2.variants.edit.flux2_klein_edit_helpers import _Flux2KleinEditHelpers
from mflux.models.flux2.weights.flux2_lora_mapping import Flux2LoRAMapping
from mflux.utils.version_util import VersionUtil


class Flux2BaseTrainingAdapter(TrainingAdapter):
    def __init__(
        self,
        *,
        model_config: ModelConfig,
        quantize: int | None,
        model_factory: Callable[[ModelConfig, int | None], Any],
    ):
        self._model_config = model_config
        self._flux2 = model_factory(model_config, quantize)
        self._guidance: float = 1.0

    def model(self):
        return self._flux2

    def transformer(self):
        return self._flux2.transformer

    def create_config(self, training_spec: TrainingSpec, *, width: int, height: int) -> Config:
        self._guidance = training_spec.guidance
        return Config(
            model_config=self._model_config,
            num_inference_steps=training_spec.steps,
            width=width,
            height=height,
            guidance=training_spec.guidance,
            scheduler="flow_match_euler_discrete",
        )

    def freeze_base(self) -> None:
        self._flux2.vae.freeze()
        self._flux2.transformer.freeze()
        self._flux2.text_encoder.freeze()

    def _encode_output_latents(self, *, image_path: Path, width: int, height: int) -> tuple[mx.array, mx.array]:
        encoded = LatentCreator.encode_image(
            vae=self._flux2.vae,
            image_path=image_path,
            height=height,
            width=width,
            tiling_config=self._flux2.tiling_config,
        )
        encoded = _Flux2KleinEditHelpers.ensure_4d_latents(encoded)
        encoded = _Flux2KleinEditHelpers.crop_to_even_spatial(encoded)
        encoded = Flux2LatentCreator.patchify_latents(encoded)
        encoded = _Flux2KleinEditHelpers.bn_normalize_vae_encoded_latents(encoded, vae=self._flux2.vae)

        clean_latents = Flux2LatentCreator.pack_latents(encoded)
        img_ids = Flux2LatentCreator.prepare_grid_ids(encoded, t_coord=0)
        return clean_latents, img_ids

    def _encode_prompt(self, *, prompt: str) -> tuple[mx.array, mx.array]:
        return Flux2PromptEncoder.encode_prompt(
            prompt=prompt,
            tokenizer=self._flux2.tokenizers["qwen3"],
            text_encoder=self._flux2.text_encoder,
            num_images_per_prompt=1,
            max_sequence_length=512,
            text_encoder_out_layers=(9, 18, 27),
        )

    def save_lora_adapter(self, *, path: Path, training_spec: TrainingSpec) -> None:  # noqa: ARG002
        weights: dict[str, mx.array] = {}
        for target in training_spec.lora_layers.targets:
            module_path = target.module_path
            if target.blocks is not None:
                for b in target.blocks.get_blocks():
                    module_path = target.module_path.format(block=b)
                    self._append_train_lora_weights(weights, self._flux2.transformer, module_path)
            else:
                self._append_train_lora_weights(weights, self._flux2.transformer, module_path)

        mx.save_safetensors(
            str(path),
            weights,
            metadata={
                "mflux_version": VersionUtil.get_mflux_version(),
                "model": training_spec.model,
            },
        )

    def load_lora_adapter(self, *, path: str | Path) -> None:
        LoRALoader.load_and_apply_lora(
            lora_mapping=Flux2LoRAMapping.get_mapping(),
            transformer=self._flux2.transformer,
            lora_paths=[str(path)],
            lora_scales=[1.0],
            role="train",
        )

    def load_training_adapter(self, *, path: str | Path, scale: float = 1.0) -> None:
        LoRALoader.load_and_apply_lora(
            lora_mapping=Flux2LoRAMapping.get_mapping(),
            transformer=self._flux2.transformer,
            lora_paths=[str(path)],
            lora_scales=[float(scale)],
            role="assistant",
        )

    def _assistant_disabled(self):
        return TrainingUtil.assistant_disabled(self._flux2.transformer)

    @staticmethod
    def _append_train_lora_weights(weights: dict[str, mx.array], transformer, module_path: str) -> None:
        train_lora = TrainingUtil.get_train_lora(transformer, module_path)

        weights[f"transformer.{module_path}.lora_A.weight"] = mx.transpose(train_lora.lora_A)
        weights[f"transformer.{module_path}.lora_B.weight"] = mx.transpose(train_lora.lora_B)



from __future__ import annotations

import datetime
import json
import os
import tempfile
import zipfile
from pathlib import Path
from typing import Protocol

from mflux.models.common.training.dataset.iterator import Iterator
from mflux.models.common.training.optimization.optimizer import Optimizer
from mflux.models.common.training.state.training_spec import (
    TRAINING_FILE_NAME_RUN_MANIFEST,
    TrainingSpec,
)
from mflux.models.common.training.state.zip_util import ZipUtil
from mflux.models.common.training.statistics.statistics import Statistics


class TrainingAdapterForCheckpointing(Protocol):
    def save_lora_adapter(self, *, path: Path, training_spec: TrainingSpec) -> None: ...


TRAINING_PATH_CHECKPOINTS = "checkpoints"
TRAINING_FILE_NAME_CHECKPOINT = "checkpoint"

TRAINING_FILE_NAME_LORA_ADAPTER = "adapter"
TRAINING_FILE_NAME_OPTIMIZER = "optimizer"
TRAINING_FILE_NAME_ITERATOR = "iterator"
TRAINING_FILE_NAME_LOSS_FILE = "loss"
TRAINING_FILE_NAME_CONFIG_FILE = "config"

TRAINING_PATH_PREVIEW_IMAGES = "preview/"
TRAINING_FILE_NAME_PREVIEW_IMAGE = "preview_image"

TRAINING_PATH_VALIDATION_PLOT = "loss/"
TRAINING_FILE_NAME_VALIDATION_LOSS = "loss"


class TrainingState:
    def __init__(self, iterator: Iterator, optimizer: Optimizer, statistics: Statistics):
        self.iterator = iterator
        self.optimizer = optimizer
        self.statistics = statistics

    def save(self, adapter: TrainingAdapterForCheckpointing, training_spec: TrainingSpec) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            optimizer_path = Path(temp_dir) / f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_OPTIMIZER}.safetensors"  # fmt: off
            lora_path = Path(temp_dir) / f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_LORA_ADAPTER}.safetensors"  # fmt: off
            iterator_path = Path(temp_dir) / f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_ITERATOR}.json"
            loss_path = Path(temp_dir) / f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_LOSS_FILE}.json"
            config_path = Path(temp_dir) / f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_CONFIG_FILE}.json"
            checkpoint_path = Path(temp_dir) / f"{TRAINING_FILE_NAME_CHECKPOINT}.json"
            manifest_path = Path(temp_dir) / TRAINING_FILE_NAME_RUN_MANIFEST
            paths = [optimizer_path, lora_path, iterator_path, loss_path, config_path, checkpoint_path, manifest_path]

            self.optimizer.save(optimizer_path)
            adapter.save_lora_adapter(path=lora_path, training_spec=training_spec)
            self.iterator.save(iterator_path)
            self.statistics.save(loss_path)
            self._save_train_config(config_path, training_spec)

            checkpoint_data = self._create_checkpoint_data(training_spec, self.iterator.start_date_time)
            with open(checkpoint_path, "w") as json_file:
                json.dump(checkpoint_data, json_file, indent=4)

            output_path = Path(training_spec.checkpoint.output_path) / TRAINING_PATH_CHECKPOINTS
            output_path.mkdir(parents=True, exist_ok=True)
            zip_path = output_path / f"{self.iterator.num_iterations:07d}_checkpoint.zip"

            run_manifest = self._create_run_manifest(
                training_spec=training_spec,
                checkpoint_data=checkpoint_data,
                checkpoint_dir=output_path,
            )
            with open(manifest_path, "w", encoding="utf-8") as json_file:
                json.dump(run_manifest, json_file, indent=4)

            with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                for file_path in paths:
                    if file_path.exists():
                        zipf.write(file_path, file_path.name)

    def _create_run_manifest(
        self,
        *,
        training_spec: TrainingSpec,
        checkpoint_data: dict,
        checkpoint_dir: Path,
    ) -> dict:
        data_root = training_spec.data_root
        relative_data_root = None
        if data_root:
            try:
                relative_data_root = os.path.relpath(data_root, checkpoint_dir)
            except ValueError:
                relative_data_root = None

        return {
            "version": 1,
            "created_at": datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "model": training_spec.model,
            "is_edit": training_spec.is_edit,
            "data_root": data_root,
            "data_root_relative": relative_data_root,
            "data_fingerprint": self._create_data_fingerprint(training_spec),
            "checkpoint_files": checkpoint_data["files"],
        }

    @staticmethod
    def _create_data_fingerprint(training_spec: TrainingSpec) -> dict:
        data_root = training_spec.data_root
        images: list[str] = []
        input_images: list[str] = []
        if data_root:
            root_path = Path(data_root).resolve()
            for item in training_spec.data:
                resolved_path = item.image.resolve()
                if resolved_path.is_relative_to(root_path):
                    images.append(str(resolved_path.relative_to(root_path)))
                else:
                    images.append(resolved_path.name)
                if item.input_image is not None:
                    resolved_input = item.input_image.resolve()
                    if resolved_input.is_relative_to(root_path):
                        input_images.append(str(resolved_input.relative_to(root_path)))
                    else:
                        input_images.append(resolved_input.name)

        return {
            "count": len(training_spec.data),
            "images": sorted(images),
            "input_images": sorted(input_images),
            "is_edit": training_spec.is_edit,
        }

    def _create_checkpoint_data(self, training_spec: TrainingSpec, start_date_time: datetime.datetime) -> dict:
        now = datetime.datetime.now()
        return {
            "metadata": {
                "start": start_date_time.strftime("%Y-%m-%d %H:%M:%S"),
                "end": now.strftime("%Y-%m-%d %H:%M:%S"),
                "duration": self._format_duration(start_date_time, now),
                "number_of_training_data": self.iterator.dataset.size(),
            },
            "files": {
                "config": f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_CONFIG_FILE}.json",
                "optimizer": f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_OPTIMIZER}.safetensors",
                "lora_adapter": f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_LORA_ADAPTER}.safetensors",
                "iterator": f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_ITERATOR}.json",
                "loss": f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_LOSS_FILE}.json",
            },
        }

    def should_save(self, training_spec: TrainingSpec) -> bool:
        return self.iterator.num_iterations % training_spec.checkpoint.save_frequency == 0

    def should_plot_loss(self, training_spec: TrainingSpec) -> bool:
        if training_spec.monitoring is None:
            return False
        return self.iterator.num_iterations % training_spec.monitoring.plot_frequency == 0

    def should_generate_image(self, training_spec: TrainingSpec) -> bool:
        if training_spec.monitoring is None:
            return False
        return self.iterator.num_iterations % training_spec.monitoring.generate_image_frequency == 0

    def get_current_preview_image_path(
        self,
        training_spec: TrainingSpec,
        *,
        preview_index: int,
        preview_name: str | None = None,
    ) -> Path:
        output_path = Path(training_spec.checkpoint.output_path) / TRAINING_PATH_PREVIEW_IMAGES
        output_path.mkdir(parents=True, exist_ok=True)
        suffix = preview_name or f"{preview_index + 1:02d}"
        return output_path / Path(f"{self.iterator.num_iterations:07d}_{TRAINING_FILE_NAME_PREVIEW_IMAGE}_{suffix}.png")

    def get_current_loss_plot_path(self, training_spec: TrainingSpec) -> Path:
        output_path = Path(training_spec.checkpoint.output_path) / TRAINING_PATH_VALIDATION_PLOT
        output_path.mkdir(parents=True, exist_ok=True)
        return output_path / Path(f"{TRAINING_FILE_NAME_VALIDATION_LOSS}.html")

    @staticmethod
    def _format_duration(start: datetime.datetime, end: datetime.datetime) -> str:
        duration = end - start
        total_seconds = int(duration.total_seconds())
        hours, remainder = divmod(total_seconds, 3600)
        minutes, seconds = divmod(remainder, 60)
        parts = []
        if hours > 0:
            parts.append(f"{hours} hour{'s' if hours > 1 else ''}")
        if minutes > 0:
            parts.append(f"{minutes} minute{'s' if minutes > 1 else ''}")
        if seconds > 0 or not parts:
            parts.append(f"{seconds} second{'s' if seconds > 1 else ''}")
        return " and ".join(parts)

    @staticmethod
    def _save_train_config(path: Path, training_spec: TrainingSpec) -> None:
        if training_spec.config_path is not None:
            with open(Path(training_spec.config_path), "r") as f:
                data = json.load(f)
                # Make data path absolute inside checkpoint config.
                data_conf = data["data"]
                if isinstance(data_conf, str):
                    resolved_path = Path(training_spec.config_path).parent / data_conf
                elif isinstance(data_conf, dict) and "path" in data_conf:
                    resolved_path = Path(training_spec.config_path).parent / str(data_conf["path"])
                else:
                    raise ValueError("Config 'data' must be a string or {'path': ...}.")
                data["data"] = str(resolved_path)
        else:
            checkpoint = ZipUtil.unzip(
                training_spec.checkpoint_path, "checkpoint.json", lambda x: json.load(open(x, "r"))
            )
            data = ZipUtil.unzip(
                training_spec.checkpoint_path, checkpoint["files"]["config"], lambda x: json.load(open(x, "r"))
            )
            if training_spec.data_root is not None:
                data["data"] = training_spec.data_root
        with open(path, "w", encoding="utf-8") as file:
            json.dump(data, file, indent=4)



from typing import TYPE_CHECKING

import mlx.core as mx

if TYPE_CHECKING:
    from mflux.models.common.config.config import Config

from mflux.models.common.schedulers.base_scheduler import BaseScheduler


class LinearScheduler(BaseScheduler):
    def __init__(self, config: "Config"):
        self.config = config
        self._sigmas = self._get_sigmas()
        self._timesteps = self._get_timesteps()

    @property
    def sigmas(self) -> mx.array:
        return self._sigmas

    @property
    def timesteps(self) -> mx.array:
        return self._timesteps

    def _get_sigmas(self) -> mx.array:
        model_config = self.config.model_config
        num_steps = self.config.num_inference_steps
        sigmas = mx.linspace(1.0, 1.0 / num_steps, num_steps)
        sigmas = mx.array(sigmas).astype(mx.float32)
        sigmas = mx.concatenate([sigmas, mx.zeros(1)])
        if model_config.requires_sigma_shift:
            m = (model_config.sigma_max_shift - model_config.sigma_base_shift) / (
                model_config.sigma_max_seq_len - model_config.sigma_base_seq_len
            )
            b = model_config.sigma_base_shift - m * model_config.sigma_base_seq_len
            mu = m * self.config.width * self.config.height / 256 + b
            mu = mx.array(mu)

            shifted = mx.exp(mu) / (mx.exp(mu) + (1 / sigmas[:-1] - 1))

            if model_config.sigma_shift_terminal is not None:
                one_minus = 1.0 - shifted
                scale = one_minus[-1] / (1.0 - model_config.sigma_shift_terminal)
                shifted = 1.0 - (one_minus / scale)

            return mx.concatenate([shifted, mx.zeros(1)])
        else:
            return sigmas

    def _get_timesteps(self) -> mx.array:
        num_steps = self.config.num_inference_steps
        timesteps = mx.arange(num_steps, dtype=mx.float32)

        return timesteps

    def step(self, noise: mx.array, timestep: int, latents: mx.array, **kwargs) -> mx.array:
        dt = (self._sigmas[timestep + 1] - self._sigmas[timestep]).astype(latents.dtype)
        return latents + noise.astype(latents.dtype) * dt



import gc
import json
from pathlib import Path

import mlx.core as mx
from PIL import Image

from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.fibo.latent_creator.fibo_latent_creator import FiboLatentCreator
from mflux.models.fibo.model.fibo_vae.wan_2_2_vae import Wan2_2_VAE
from mflux.models.fibo_vlm.model.fibo_vlm import FiboVLM
from mflux.utils.image_util import ImageUtil
from mflux.utils.prompt_util import PromptUtil

FIBO_EDIT_RMBG_DEFAULT_EDIT_INSTRUCTION = (
    "Generate a detailed grayscale alpha matte. Map the opaque foreground to white "
    "and the background to black. Produce soft, anti-aliased grayscale gradients at the "
    "edges of the subject to represent fine details and transparency."
)
FIBO_EDIT_RMBG_DEFAULT_JSON_PROMPT = json.dumps({"edit_instruction": FIBO_EDIT_RMBG_DEFAULT_EDIT_INSTRUCTION})


class FiboEditUtil:
    @staticmethod
    def get_json_prompt_for_edit(
        args,
        quantize: int | None,
        default_json_prompt_if_missing: str | None = None,
    ) -> str:
        prompt = PromptUtil.read_prompt(args)
        missing = prompt is None or (isinstance(prompt, str) and not prompt.strip())
        if missing:
            if default_json_prompt_if_missing is not None:
                return FiboEditUtil.ensure_edit_instruction(default_json_prompt_if_missing)
            raise ValueError(
                "FIBO edit requires an edit instruction via --prompt/--prompt-file, or a JSON prompt from metadata."
            )

        try:
            return FiboEditUtil.ensure_edit_instruction(prompt)
        except (TypeError, ValueError):
            pass

        if getattr(args, "image_path", None) is None:
            raise ValueError("Edit mode requires --image-path when prompt is not valid JSON.")

        image = ImageUtil.load_image(args.image_path)
        if getattr(args, "mask_path", None) is not None:
            mask_image = Image.open(args.mask_path).convert("L")
            if mask_image.size != image.size:
                raise ValueError("Mask and image must have the same size.")
            image = FiboEditUtil._composite_mask_on_image(mask=mask_image, image=image)

        vlm = FiboVLM(quantize=quantize)
        try:
            return vlm.edit(
                image=image,
                edit_instruction=prompt,
                use_mask=getattr(args, "mask_path", None) is not None,
                seed=42,
            )
        finally:
            del vlm
            gc.collect()
            mx.clear_cache()

    @staticmethod
    def parse_json_prompt(prompt: str) -> dict:
        try:
            value = json.loads(prompt)
        except (TypeError, json.JSONDecodeError) as exc:
            raise ValueError("FIBO edit prompt must be a valid JSON string.") from exc

        if not isinstance(value, dict):
            raise ValueError("FIBO edit prompt JSON must be an object.")
        return value

    @staticmethod
    def ensure_edit_instruction(prompt: str, edit_instruction: str | None = None) -> str:
        prompt_dict = FiboEditUtil.parse_json_prompt(prompt)
        if "edit_instruction" in prompt_dict and prompt_dict["edit_instruction"]:
            return json.dumps(prompt_dict)

        if edit_instruction is None or not edit_instruction.strip():
            raise ValueError("FIBO edit prompt JSON must include `edit_instruction`.")

        prompt_dict["edit_instruction"] = edit_instruction.strip()
        return json.dumps(prompt_dict)

    @staticmethod
    def load_edit_image(
        image_path: Path | str,
        width: int,
        height: int,
        mask_path: Path | str | None = None,
    ) -> Image.Image:
        image = ImageUtil.load_image(image_path)
        if mask_path is None:
            return ImageUtil.scale_to_dimensions(image, width, height)

        mask_image = Image.open(mask_path).convert("L")
        if mask_image.size != image.size:
            raise ValueError("Mask and image must have the same size.")

        masked_image = FiboEditUtil._composite_mask_on_image(mask=mask_image, image=image)
        return ImageUtil.scale_to_dimensions(masked_image, width, height)

    @staticmethod
    def encode_conditioning_image(
        vae: Wan2_2_VAE,
        image: Image.Image,
        height: int,
        width: int,
        tiling_config=None,
    ) -> mx.array:
        image_array = ImageUtil.to_array(image=image)
        image_latents = VAEUtil.encode(vae=vae, image=image_array, tiling_config=tiling_config)
        return FiboLatentCreator.pack_latents(latents=image_latents, height=height, width=width)

    @staticmethod
    def create_conditioning_image_ids(height: int, width: int, dtype: mx.Dtype) -> mx.array:
        latent_height = height // 16
        latent_width = width // 16
        row_indices = mx.arange(0, latent_height, dtype=dtype)[:, None]
        row_indices = mx.broadcast_to(row_indices, (latent_height, latent_width))
        col_indices = mx.arange(0, latent_width, dtype=dtype)[None, :]
        col_indices = mx.broadcast_to(col_indices, (latent_height, latent_width))
        ones_channel = mx.ones((latent_height, latent_width), dtype=dtype)
        latent_image_ids = mx.stack([ones_channel, row_indices, col_indices], axis=-1)
        latent_image_ids = mx.reshape(latent_image_ids, (1, latent_height * latent_width, 3))
        return latent_image_ids

    @staticmethod
    def _composite_mask_on_image(mask: Image.Image, image: Image.Image) -> Image.Image:
        gray_img = Image.new("RGB", image.size, (128, 128, 128))
        return Image.composite(gray_img, image.convert("RGB"), mask.convert("L"))

    @staticmethod
    def build_rgba_composite_image(source_image_path: Path | str, matte_image: Image.Image) -> Image.Image:
        base = Image.open(source_image_path).convert("RGB").copy()
        alpha = matte_image.convert("L").resize(base.size, Image.LANCZOS)
        base.putalpha(alpha)
        return base



from __future__ import annotations

import gc
import random
import tempfile
from pathlib import Path

import mlx.core as mx
from mlx import nn
from mlx.utils import tree_unflatten
from PIL import Image as PILImage
from tqdm import tqdm

from mflux.models.common.latent_creator.latent_creator import LatentCreator
from mflux.models.common.lora.layer.fused_linear_lora_layer import FusedLoRALinear
from mflux.models.common.lora.layer.linear_lora_layer import LoRALinear
from mflux.models.common.training.adapters.base import TrainingAdapter
from mflux.models.common.training.dataset.batch import Batch, DataItem
from mflux.models.common.training.state.training_spec import TrainingSpec
from mflux.models.common.training.state.training_state import TrainingState
from mflux.models.common.training.statistics.plotter import Plotter
from mflux.models.common.training.utils import TrainingUtil


class TrainingTrainer:
    @staticmethod
    def compute_loss(
        adapter: TrainingAdapter,
        training_spec: TrainingSpec,
        base_config,
        batch: Batch,
    ) -> mx.float16:
        losses = [
            TrainingTrainer._single_example_loss(adapter, training_spec, base_config, item, batch.rng)
            for item in batch.data
        ]
        return mx.mean(mx.array(losses))

    @staticmethod
    def _single_example_loss(
        adapter: TrainingAdapter,
        training_spec: TrainingSpec,
        base_config,
        item: DataItem,
        rng: random.Random,
    ) -> mx.float16:
        # Create a config matching this item's spatial dimensions.
        # Flux uses config.width/height for rotary embeddings, so this must match the latent layout.
        config = adapter.create_config(training_spec, width=item.width, height=item.height)

        # Reuse the base scheduler only when compatible with the item's dimensions.
        # Some schedulers depend on image seq len when sigma shift is enabled.
        if not config.model_config.requires_sigma_shift or config.image_seq_len == base_config.image_seq_len:
            config._scheduler = base_config.scheduler  # type: ignore[attr-defined]
        else:
            _ = config.scheduler

        time_seed = rng.randint(0, 2**32 - 1)
        noise_seed = rng.randint(0, 2**32 - 1)

        low = int(training_spec.training_loop.timestep_low)
        high = int(
            config.num_inference_steps
            if training_spec.training_loop.timestep_high is None
            else training_spec.training_loop.timestep_high
        )

        t = int(
            mx.random.randint(
                low=low,
                high=high,
                shape=[],
                key=mx.random.key(time_seed),
            )
        )

        clean_image = item.clean_latents
        pure_noise = mx.random.normal(
            shape=clean_image.shape,
            dtype=config.precision,
            key=mx.random.key(noise_seed),
        )

        latents_t = LatentCreator.add_noise_by_interpolation(
            clean=clean_image,
            noise=pure_noise,
            sigma=config.scheduler.sigmas[t],
        )

        predicted_noise = adapter.predict_noise(
            t=t,
            latents_t=latents_t,
            sigmas=config.scheduler.sigmas,
            cond=item.cond,
            config=config,
        )

        error = (clean_image + predicted_noise - pure_noise).square()
        return error.mean()

    @staticmethod
    def train(
        *,
        adapter: TrainingAdapter,
        training_spec: TrainingSpec,
        training_state: TrainingState,
    ) -> None:
        first_preview = None
        if training_spec.monitoring is not None and training_spec.monitoring.preview_images:
            first_preview = training_spec.monitoring.preview_images[0]
        preview_width, preview_height = TrainingTrainer._preview_dimensions(training_spec, preview_image=first_preview)
        base_config = adapter.create_config(training_spec, width=preview_width, height=preview_height)
        # Ensure scheduler is initialized once and can be reused in per-item configs.
        _ = base_config.scheduler

        # Freeze base weights and unfreeze LoRA weights
        adapter.freeze_base()
        TrainingTrainer._unfreeze_lora_layers(adapter.transformer())

        train_step_function = nn.value_and_grad(
            model=adapter.model(),
            fn=lambda b: TrainingTrainer.compute_loss(adapter, training_spec, base_config, b),
        )

        if training_spec.monitoring is not None and training_state.iterator.num_iterations == 0:
            TrainingTrainer._generate_previews_with_optimizer_offload(adapter, training_spec, training_state)
            validation_batch = training_state.iterator.get_validation_batch()
            validation_loss = TrainingTrainer.compute_loss(adapter, training_spec, base_config, validation_batch)
            training_state.statistics.append_values(step=training_state.iterator.num_iterations, loss=float(validation_loss))  # fmt: off
            Plotter.update_loss_plot(training_spec=training_spec, training_state=training_state)
            del validation_loss
            training_state.save(adapter, training_spec)

        batches = tqdm(
            training_state.iterator,
            total=training_state.iterator.total_number_of_steps(),
            initial=training_state.iterator.num_iterations,
        )

        for batch in batches:
            loss, grads = train_step_function(batch)
            training_state.optimizer.optimizer.update(model=adapter.model(), gradients=grads)
            mx.eval(adapter.model().parameters(), training_state.optimizer.optimizer.state)
            del loss, grads

            if training_state.should_plot_loss(training_spec):
                validation_batch = training_state.iterator.get_validation_batch()
                validation_loss = TrainingTrainer.compute_loss(adapter, training_spec, base_config, validation_batch)
                training_state.statistics.append_values(step=training_state.iterator.num_iterations, loss=float(validation_loss))  # fmt: off
                Plotter.update_loss_plot(training_spec=training_spec, training_state=training_state)
                del validation_loss

            if training_state.should_generate_image(training_spec):
                TrainingTrainer._generate_previews_with_optimizer_offload(adapter, training_spec, training_state)

            if training_state.should_save(training_spec):
                training_state.save(adapter, training_spec)

            if training_spec.low_ram:
                mx.clear_cache()

        training_state.save(adapter, training_spec)

    @staticmethod
    def _unfreeze_lora_layers(module: nn.Module) -> None:
        for _, child in module.named_modules():
            if isinstance(child, LoRALinear):
                if getattr(child, "_mflux_lora_role", None) == "train":
                    child.unfreeze(keys=["lora_A", "lora_B"], strict=False)
            elif isinstance(child, FusedLoRALinear):
                for lora in child.loras:
                    if getattr(lora, "_mflux_lora_role", None) == "train":
                        lora.unfreeze(keys=["lora_A", "lora_B"], strict=False)

    @staticmethod
    def _preview_dimensions(training_spec: TrainingSpec, *, preview_image: Path | None = None) -> tuple[int, int]:
        if training_spec.monitoring is None:
            return 1024, 1024
        if preview_image is not None:
            with PILImage.open(preview_image) as img:
                width, height = img.size
        else:
            width = int(training_spec.monitoring.preview_width)
            height = int(training_spec.monitoring.preview_height)

        return TrainingUtil.resolve_dimensions(
            width=width,
            height=height,
            max_resolution=None,
            error_template=(
                f"Preview image too small for training (needs >=16px). Got {{width}}x{{height}} from {preview_image}"
            ),
        )

    @staticmethod
    def _generate_previews(
        adapter: TrainingAdapter,
        training_spec: TrainingSpec,
        training_state: TrainingState,
    ) -> None:
        if training_spec.monitoring is None:
            return
        preview_prompts = training_spec.monitoring.preview_prompts
        preview_names = training_spec.monitoring.preview_prompt_names
        preview_images = training_spec.monitoring.preview_images
        for idx, prompt in enumerate(preview_prompts):
            image_paths = None
            if training_spec.is_edit:
                if not preview_images or idx >= len(preview_images):
                    raise ValueError("Edit training requires data/preview.* for each preview prompt.")
                image_paths = [preview_images[idx]]
                preview_width, preview_height = TrainingTrainer._preview_dimensions(
                    training_spec, preview_image=preview_images[idx]
                )
            else:
                preview_width, preview_height = TrainingTrainer._preview_dimensions(training_spec)
            image = adapter.generate_preview_image(
                seed=training_spec.seed,
                prompt=prompt,
                width=preview_width,
                height=preview_height,
                steps=training_spec.steps,
                image_paths=image_paths,
            )
            preview_name = preview_names[idx] if idx < len(preview_names) else None
            image.save(
                training_state.get_current_preview_image_path(
                    training_spec,
                    preview_index=idx,
                    preview_name=preview_name,
                )
            )
            del image

    @staticmethod
    def _generate_previews_with_optimizer_offload(
        adapter: TrainingAdapter,
        training_spec: TrainingSpec,
        training_state: TrainingState,
    ) -> None:
        optimizer = training_state.optimizer
        with tempfile.TemporaryDirectory() as tmp_dir:
            offload_path = Path(tmp_dir) / "optimizer_offload.safetensors"
            optimizer.save(offload_path)
            optimizer.optimizer.state = []

            gc.collect()
            mx.clear_cache()
            try:
                TrainingTrainer._generate_previews(adapter, training_spec, training_state)
            finally:
                restored_state = tree_unflatten(list(mx.load(str(offload_path)).items()))
                optimizer.optimizer.state = restored_state
                gc.collect()
                mx.clear_cache()



from pathlib import Path

import mlx.core as mx
from mlx import nn
from PIL import Image

from mflux.models.common.config.config import Config
from mflux.models.common.config.model_config import ModelConfig
from mflux.models.common.latent_creator.latent_creator import Img2Img, LatentCreator
from mflux.models.common.vae.vae_util import VAEUtil
from mflux.models.common.weights.saving.model_saver import ModelSaver
from mflux.models.z_image.latent_creator import ZImageLatentCreator
from mflux.models.z_image.model.z_image_text_encoder.prompt_encoder import PromptEncoder
from mflux.models.z_image.model.z_image_text_encoder.text_encoder import TextEncoder
from mflux.models.z_image.model.z_image_transformer.transformer import ZImageTransformer
from mflux.models.z_image.model.z_image_vae.vae import VAE
from mflux.models.z_image.weights.z_image_weight_definition import ZImageWeightDefinition
from mflux.models.z_image.z_image_initializer import ZImageInitializer
from mflux.utils.apple_silicon import AppleSiliconUtil
from mflux.utils.exceptions import StopImageGenerationException
from mflux.utils.image_util import ImageUtil


class ZImage(nn.Module):
    vae: VAE
    text_encoder: TextEncoder
    transformer: ZImageTransformer

    def __init__(
        self,
        quantize: int | None = None,
        model_path: str | None = None,
        lora_paths: list[str] | None = None,
        lora_scales: list[float] | None = None,
        model_config: ModelConfig = ModelConfig.z_image_turbo(),
    ):
        super().__init__()
        ZImageInitializer.init(
            model=self,
            quantize=quantize,
            model_path=model_path,
            lora_paths=lora_paths,
            lora_scales=lora_scales,
            model_config=model_config,
        )

    def generate_image(
        self,
        seed: int,
        prompt: str,
        num_inference_steps: int = 4,
        height: int = 1024,
        width: int = 1024,
        guidance: float | None = None,
        image_path: Path | str | None = None,
        image_strength: float | None = None,
        scheduler: str | None = None,
        negative_prompt: str | None = None,
    ) -> Image.Image:
        supports_guidance = bool(self.model_config.supports_guidance)
        if not supports_guidance:
            guidance = 0.0

        if scheduler is None:
            scheduler = "flow_match_euler_discrete" if supports_guidance else "linear"

        # 0. Create a new config based on the model type and input parameters
        config = Config(
            width=width,
            height=height,
            guidance=guidance,
            scheduler=scheduler,
            image_path=image_path,
            image_strength=image_strength,
            model_config=self.model_config,
            num_inference_steps=num_inference_steps,
        )
        # 1. Create the initial latents
        latents = LatentCreator.create_for_txt2img_or_img2img(
            seed=seed,
            width=config.width,
            height=config.height,
            img2img=Img2Img(
                vae=self.vae,
                latent_creator=ZImageLatentCreator,
                image_path=config.image_path,
                sigmas=config.scheduler.sigmas,
                init_time_step=config.init_time_step,
                tiling_config=self.tiling_config,
            ),
        )
        text_encodings, negative_encodings = self._encode_prompts(
            prompt=prompt,
            negative_prompt=negative_prompt,
            guidance=config.guidance,
        )

        # 3. Create callback context and call before_loop
        ctx = self.callbacks.start(seed=seed, prompt=prompt, config=config)
        ctx.before_loop(latents)
        predict = self._predict(self.transformer)

        for t in config.time_steps:
            try:
                # 4.t Predict the noise
                sigma_t = config.scheduler.sigmas[t].reshape((1,))
                timestep = mx.ones_like(sigma_t) - sigma_t
                noise = predict(
                    latents=latents,
                    timestep=timestep,
                    sigmas=config.scheduler.sigmas,
                    text_encodings=text_encodings,
                    negative_encodings=negative_encodings,
                    guidance=config.guidance,
                )

                # 5.t Take one denoise step
                latents = config.scheduler.step(noise=noise, timestep=t, latents=latents)

                # 6.t Call subscribers in-loop
                ctx.in_loop(t, latents)

                # (Optional) Evaluate to enable progress tracking
                mx.eval(latents)

            except KeyboardInterrupt:  # noqa: PERF203
                ctx.interruption(t, latents)
                raise StopImageGenerationException(
                    f"Stopping image generation at step {t + 1}/{config.num_inference_steps}"
                )

        # 7. Call subscribers after loop
        ctx.after_loop(latents)

        # 8. Decode the latents and return the image
        decoded = self._decode_latents(latents=latents, config=config)
        return ImageUtil.to_image(
            decoded_latents=decoded,
            config=config,
            seed=seed,
            prompt=prompt,
            quantization=self.bits,
            lora_paths=self.lora_paths,
            lora_scales=self.lora_scales,
            image_path=config.image_path,
            image_strength=config.image_strength,
            generation_time=config.time_steps.format_dict["elapsed"],
            negative_prompt=negative_prompt,
        )

    def _encode_prompts(
        self,
        *,
        prompt: str,
        negative_prompt: str | None,
        guidance: float,
    ) -> tuple[mx.array, mx.array | None]:
        text_encodings = PromptEncoder.encode_prompt(
            prompt=prompt,
            tokenizer=self.tokenizers["z_image"],
            text_encoder=self.text_encoder,
        )
        if guidance <= 1.0:
            return text_encodings, None
        negative_text = negative_prompt if negative_prompt and negative_prompt.strip() else " "
        negative_encodings = PromptEncoder.encode_prompt(
            prompt=negative_text,
            tokenizer=self.tokenizers["z_image"],
            text_encoder=self.text_encoder,
        )
        return text_encodings, negative_encodings

    def _decode_latents(self, *, latents: mx.array, config: Config) -> mx.array:
        unpacked = ZImageLatentCreator.unpack_latents(latents, config.height, config.width)
        return VAEUtil.decode(vae=self.vae, latent=unpacked, tiling_config=self.tiling_config)

    def save_model(self, base_path: str) -> None:
        ModelSaver.save_model(
            model=self,
            bits=self.bits,
            base_path=base_path,
            weight_definition=ZImageWeightDefinition,
        )

    @staticmethod
    def _predict(transformer: ZImageTransformer):
        def predict(
            latents: mx.array,
            timestep: mx.array,
            sigmas: mx.array,
            text_encodings: mx.array,
            negative_encodings: mx.array | None,
            guidance: float,
        ) -> mx.array:
            noise = transformer(
                timestep=timestep,
                x=latents,
                cap_feats=text_encodings,
                sigmas=sigmas,
            )
            if negative_encodings is None:
                return noise
            negative_noise = transformer(
                timestep=timestep,
                x=latents,
                cap_feats=negative_encodings,
                sigmas=sigmas,
            )
            return noise + guidance * (noise - negative_noise)

        if AppleSiliconUtil.is_m1_or_m2():
            return predict
        return mx.compile(predict)



import os
from pathlib import Path

import platformdirs

BATTERY_PERCENTAGE_STOP_LIMIT = 5
CONTROLNET_STRENGTH = 0.4
DEFAULT_DEV_FILL_GUIDANCE = 30
DEFAULT_DEPTH_GUIDANCE = 10
DIMENSION_STEP_PIXELS = 16
GUIDANCE_SCALE = 3.5
GUIDANCE_SCALE_KONTEXT = 2.5
HEIGHT, WIDTH = 1024, 1024
IMAGE_STRENGTH = 0.4
MODEL_CHOICES = [
    "dev",
    "schnell",
    "krea-dev",
    "dev-krea",
    "qwen",
    "fibo",
    "fibo-lite",
    "fibo-edit",
    "fibo-edit-rmbg",
    "z-image",
    "z-image-turbo",
    "flux2-klein-4b",
    "flux2-klein-9b",
    "flux2-klein-base-4b",
    "flux2-klein-base-9b",
]
MODEL_INFERENCE_STEPS = {
    "dev": 25,
    "schnell": 4,
    "krea-dev": 25,
    "qwen": 20,
    "qwen-image": 20,
    "qwen-image-edit": 20,
    "fibo": 50,
    "fibo-lite": 8,
    "fibo-edit": 50,
    "fibo-edit-rmbg": 10,
    "z-image": 50,
    "z-image-turbo": 9,
    "flux2-klein-4b": 4,
    "flux2-klein-9b": 4,
    "flux2-klein-base-4b": 50,
    "flux2-klein-base-9b": 50,
}
QUANTIZE_CHOICES = [3, 5, 4, 6, 8]

if os.environ.get("MFLUX_CACHE_DIR"):
    MFLUX_CACHE_DIR = Path(os.environ["MFLUX_CACHE_DIR"]).resolve()
else:
    MFLUX_CACHE_DIR = Path(platformdirs.user_cache_dir(appname="mflux"))

MFLUX_LORA_CACHE_DIR = MFLUX_CACHE_DIR / "loras"



import mlx.core as mx
from mlx import nn

from mflux.models.flux2.model.flux2_vae.common.attention import Flux2AttentionBlock
from mflux.models.flux2.model.flux2_vae.common.resnet_block_2d import Flux2ResnetBlock2D


class Flux2UNetMidBlock2D(nn.Module):
    def __init__(self, channels: int, eps: float = 1e-6, groups: int = 32, add_attention: bool = True):
        super().__init__()
        self.resnets = [
            Flux2ResnetBlock2D(channels, channels, eps=eps, groups=groups),
            Flux2ResnetBlock2D(channels, channels, eps=eps, groups=groups),
        ]
        self.attentions = [Flux2AttentionBlock(channels, groups=groups, eps=eps)] if add_attention else []

    def __call__(self, hidden_states: mx.array) -> mx.array:
        hidden_states = self.resnets[0](hidden_states)
        if self.attentions:
            hidden_states = self.attentions[0](hidden_states)
        hidden_states = self.resnets[1](hidden_states)
        return hidden_states
